from datetime import timedelta, timezone, tzinfo
from zoneinfo import ZoneInfo

import pytest

from allotropy.allotrope.allotrope import AllotropeConversionError
from allotropy.parsers.utils.timestamp_parser import TimestampParser


def test_timestamp_parser_init_fails_invalid_default_timezone() -> None:
    with pytest.raises(
        AllotropeConversionError, match="Invalid default timezone 'timezone'"
    ):
        TimestampParser("timezone")  # type: ignore[arg-type]


@pytest.mark.short
@pytest.mark.parametrize(
    "time_str,expected",
    [
        ("10-11-08", "2008-10-11T00:00:00+00:00"),
        ("Fri, 11 Nov 2011 03:18:09", "2011-11-11T03:18:09+00:00"),
        ("Fri, 11 Nov 2011 03:18:09 -0400", "2011-11-11T03:18:09-04:00"),
        ("Tue Jun 22 07:46:22 EST 2010", "2010-06-22T07:46:22-05:00"),
        ("Tue Jun 22 07:46:22 EDT 2010", "2010-06-22T07:46:22-04:00"),
        ("Tue Jun 22 07:46:22 GMT 2010", "2010-06-22T07:46:22+00:00"),
    ],
)
def test_timestamp_parser_default_utc(time_str: str, expected: str) -> None:
    assert TimestampParser().parse(time_str) == expected


# Similar timezones, but different DST info.
US_PACIFIC = ZoneInfo("US/Pacific")
UTC_MINUS_7 = timezone(timedelta(hours=-7))


@pytest.mark.short
@pytest.mark.parametrize(
    "default_timezone,time_str,expected",
    [
        (US_PACIFIC, "10-11-08", "2008-10-11T00:00:00-07:00"),
        (US_PACIFIC, "Fri, 11 Nov 2011 03:18:09", "2011-11-11T03:18:09-08:00"),
        (US_PACIFIC, "Fri, 11 Jun 2011 03:18:09", "2011-06-11T03:18:09-07:00"),
        (US_PACIFIC, "Fri, 11 Nov 2011 03:18:09 -0400", "2011-11-11T03:18:09-04:00"),
        (US_PACIFIC, "Tue Jun 22 07:46:22 EST 2010", "2010-06-22T07:46:22-05:00"),
        (UTC_MINUS_7, "10-11-08", "2008-10-11T00:00:00-07:00"),
        (UTC_MINUS_7, "Fri, 11 Nov 2011 03:18:09", "2011-11-11T03:18:09-07:00"),
        (UTC_MINUS_7, "Fri, 11 Jun 2011 03:18:09", "2011-06-11T03:18:09-07:00"),
        (UTC_MINUS_7, "Fri, 11 Nov 2011 03:18:09 -0400", "2011-11-11T03:18:09-04:00"),
        (UTC_MINUS_7, "Tue Jun 22 07:46:22 EST 2010", "2010-06-22T07:46:22-05:00"),
    ],
)
def test_timestamp_parser_provided_timezone(
    default_timezone: tzinfo, time_str: str, expected: str
) -> None:
    assert TimestampParser(default_timezone).parse(time_str) == expected


@pytest.mark.short
@pytest.mark.parametrize("time_str", ["blah"])
def test_timestamp_parser_returns_none_on_invalid_timestamp(time_str: str) -> None:
    assert TimestampParser().parse(time_str) is None


@pytest.mark.short
def test_timestamp_parser_handles_24h_pm() -> None:
    assert (
        TimestampParser().parse("2023-03-16 16:52:37 PM") == "2023-03-16T16:52:37+00:00"
    )



# generated by datamodel-codegen:
#   filename:  cell-culture-analyzer.json
#   timestamp: 2023-09-21T17:20:28+00:00

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Union

from allotropy.allotrope.models.shared.definitions.custom import (
    TNullableQuantityValueCell,
    TNullableQuantityValueGramPerLiter,
    TNullableQuantityValueMicrometer,
    TNullableQuantityValueMillimeterOfMercury,
    TNullableQuantityValueMillimolePerLiter,
    TNullableQuantityValueMillionCellsPerMilliliter,
    TNullableQuantityValueMilliOsmolesPerKilogram,
    TNullableQuantityValueOpticalDensity,
    TNullableQuantityValuePercent,
    TNullableQuantityValuePH,
    TNullableQuantityValueTODO,
    TNullableQuantityValueUnitPerLiter,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TDateTimeValue,
    TStringValue,
)


@dataclass
class DeviceSystemDocument:
    device_identifier: Optional[TStringValue] = None
    model_number: Optional[TStringValue] = None
    device_serial_number: Optional[TStringValue] = None


@dataclass
class SampleDocument:
    sample_identifier: Optional[TStringValue] = None
    batch_identifier: Optional[TStringValue] = None
    sample_role_type: Optional[str] = None


@dataclass
class AnalyteDocumentItem:
    analyte_name: TStringValue
    molar_concentration: Optional[
        Union[
            TNullableQuantityValueMillimolePerLiter,
            TNullableQuantityValueGramPerLiter,
            TNullableQuantityValueUnitPerLiter,
        ]
    ] = None


@dataclass
class AnalyteAggregateDocument:
    analyte_document: Optional[list[AnalyteDocumentItem]] = None


@dataclass
class MeasurementDocumentItem:
    sample_document: SampleDocument
    measurement_time: TDateTimeValue
    analyte_aggregate_document: AnalyteAggregateDocument
    pco2: Optional[TNullableQuantityValueMillimeterOfMercury] = None
    co2_saturation: Optional[TNullableQuantityValuePercent] = None
    po2: Optional[TNullableQuantityValueMillimeterOfMercury] = None
    o2_saturation: Optional[TNullableQuantityValuePercent] = None
    optical_density: Optional[TNullableQuantityValueOpticalDensity] = None
    pH: Optional[TNullableQuantityValuePH] = None
    osmolality: Optional[TNullableQuantityValueMilliOsmolesPerKilogram] = None
    viability__cell_counter_: Optional[TNullableQuantityValuePercent] = None
    total_cell_density__cell_counter_: Optional[
        TNullableQuantityValueMillionCellsPerMilliliter
    ] = None
    viable_cell_density__cell_counter_: Optional[
        TNullableQuantityValueMillionCellsPerMilliliter
    ] = None
    average_live_cell_diameter__cell_counter_: Optional[
        TNullableQuantityValueMicrometer
    ] = None
    average_total_cell_diameter__cell_counter_: Optional[
        TNullableQuantityValueMicrometer
    ] = None
    total_cell_diameter_distribution__cell_counter_: Optional[
        TNullableQuantityValueTODO
    ] = None
    viable_cell_count__cell_counter_: Optional[TNullableQuantityValueCell] = None
    total_cell_count__cell_counter_: Optional[TNullableQuantityValueCell] = None


@dataclass
class MeasurementAggregateDocument:
    measurement_identifier: TStringValue
    analyst: TStringValue
    device_system_document: DeviceSystemDocument
    measurement_document: list[MeasurementDocumentItem]
    data_processing_time: Optional[TDateTimeValue] = None


@dataclass
class Model:
    measurement_aggregate_document: Optional[MeasurementAggregateDocument] = None
    manifest: str = "http://purl.allotrope.org/manifests/cell-culture-analyzer/BENCHLING/2023/09/cell-culture-analyzer.manifest"



import pytest

from tests.parsers.roche_cedex_bioht.roche_cedex_bioht_data import (
    get_reader,
    get_reader_samples,
    get_reader_title,
)


@pytest.mark.short
def test_data_reader() -> None:
    reader = get_reader()

    assert reader.title_data.equals(get_reader_title())
    assert reader.samples_data.equals(get_reader_samples())



# header
# sample setup
# raw data
# amplification data
# multicomponent data
# results
# melt curve raw data

from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from typing import Optional

from allotropy.allotrope.allotrope import AllotropeConversionError
from allotropy.allotrope.models.pcr_benchling_2023_09_qpcr import (
    ExperimentType,
)
from allotropy.parsers.appbio_quantstudio.calculated_document import CalculatedDocument
from allotropy.parsers.appbio_quantstudio.referenceable import Referenceable


@dataclass
class Header:
    measurement_time: str
    plate_well_count: int
    experiment_type: ExperimentType
    device_identifier: str
    model_number: str
    device_serial_number: str
    measurement_method_identifier: str
    pcr_detection_chemistry: str
    passive_reference_dye_setting: Optional[str]
    barcode: Optional[str]
    analyst: Optional[str]
    experimental_data_identifier: Optional[str]


@dataclass
class WellItem(Referenceable):
    uuid: str
    identifier: int
    target_dna_description: str
    sample_identifier: str
    reporter_dye_setting: Optional[str]
    position: Optional[str]
    well_location_identifier: Optional[str]
    quencher_dye_setting: Optional[str]
    sample_role_type: Optional[str]
    amplification_data_obj: Optional[AmplificationData] = None
    result_obj: Optional[Result] = None

    def add_amplification_data(self, amplification_data: AmplificationData) -> None:
        self.amplification_data_obj = amplification_data

    @property
    def amplification_data(self) -> AmplificationData:
        if self.amplification_data_obj is None:
            msg = f"Unable to find amplification data for well {self.identifier}"
            raise AllotropeConversionError(msg)
        return self.amplification_data_obj

    def add_result(self, result: Result) -> None:
        self.result_obj = result

    @property
    def result(self) -> Result:
        if self.result_obj is None:
            msg = f"Unablt to find result data for well {self.identifier}"
            raise AllotropeConversionError(msg)
        return self.result_obj


@dataclass
class Well:
    identifier: int
    items: dict[str, WellItem]
    multicomponent_data: Optional[MulticomponentData] = None
    melt_curve_raw_data: Optional[MeltCurveRawData] = None
    calculated_document: Optional[CalculatedDocument] = None

    def get_well_item(self, target: str) -> WellItem:
        well_item = self.items.get(target)
        if well_item is None:
            msg = f"Unable to find target dna {target} for well {self.identifier}"
            raise AllotropeConversionError(msg)
        return well_item

    def get_an_well_item(self) -> Optional[WellItem]:
        if not self.items:
            return None
        target, *_ = self.items.keys()
        return self.items[target]

    def add_multicomponent_data(self, multicomponent_data: MulticomponentData) -> None:
        self.multicomponent_data = multicomponent_data

    def add_melt_curve_raw_data(self, melt_curve_raw_data: MeltCurveRawData) -> None:
        self.melt_curve_raw_data = melt_curve_raw_data


@dataclass
class WellList:
    wells: list[Well]

    def iter_well_items(self) -> Iterator[WellItem]:
        for well in self.wells:
            yield from well.items.values()

    def __iter__(self) -> Iterator[Well]:
        return iter(self.wells)


@dataclass
class RawData:
    lines: list[str]


@dataclass
class AmplificationData:
    total_cycle_number_setting: float
    cycle: list[float]
    rn: list[Optional[float]]
    delta_rn: list[Optional[float]]


@dataclass
class MulticomponentData:
    cycle: list[float]
    columns: dict[str, list[Optional[float]]]

    def get_column(self, name: str) -> list[Optional[float]]:
        column = self.columns.get(name)
        if column is None:
            msg = f"Unable to obtain {name} from multicomponent data"
            raise AllotropeConversionError(msg)
        return column


@dataclass
class Result:
    cycle_threshold_value_setting: float
    cycle_threshold_result: Optional[float]
    automatic_cycle_threshold_enabled_setting: Optional[bool]
    automatic_baseline_determination_enabled_setting: Optional[bool]
    normalized_reporter_result: Optional[float]
    baseline_corrected_reporter_result: Optional[float]
    genotyping_determination_result: Optional[str]
    genotyping_determination_method_setting: Optional[float]
    quantity: Optional[float]
    quantity_mean: Optional[float]
    quantity_sd: Optional[float]
    ct_mean: Optional[float]
    ct_sd: Optional[float]
    delta_ct_mean: Optional[float]
    delta_ct_se: Optional[float]
    delta_delta_ct: Optional[float]
    rq: Optional[float]
    rq_min: Optional[float]
    rq_max: Optional[float]
    rn_mean: Optional[float]
    rn_sd: Optional[float]
    y_intercept: Optional[float]
    r_squared: Optional[float]
    slope: Optional[float]
    efficiency: Optional[float]


@dataclass
class MeltCurveRawData:
    reading: list[float]
    fluorescence: list[Optional[float]]
    derivative: list[Optional[float]]


@dataclass
class Data:
    header: Header
    wells: WellList
    raw_data: Optional[RawData]
    endogenous_control: str
    reference_sample: str
    calculated_documents: list[CalculatedDocument]



from __future__ import annotations

from abc import ABC
from collections import defaultdict
from enum import Enum
from typing import Any, ClassVar, Optional

from allotropy.allotrope.allotrope import AllotropeConversionError
from allotropy.allotrope.models.shared.components.plate_reader import (
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
    TDatacube,
    TDatacubeComponent,
    TDatacubeData,
    TDatacubeStructure,
)
from allotropy.parsers.utils.values import assert_not_none, PrimitiveValue, try_int

START_LETTER_CODE = ord("A")
EXPORT_VERSION = "1.3"


class ReadType(Enum):
    ENDPOINT = "Endpoint"
    KINETIC = "Kinetic"
    SPECTRUM = "Spectrum"
    WELL_SCAN = "Well Scan"


class ExportFormat(Enum):
    TIME_FORMAT = "TimeFormat"
    PLATE_FORMAT = "PlateFormat"


class DataType(Enum):
    RAW = "Raw"
    REDUCED = "Reduced"


class Block:
    BLOCK_TYPE: ClassVar[str]

    def __init__(self, raw_lines: list[str]):
        self.raw_lines = raw_lines


class GroupBlock(Block):
    GROUP_PREFIX = "Group: "
    GROUP_COLUMN_PREFIX = "Group Column"
    BLOCK_TYPE = "Group"
    name: str
    group_data: list[str]

    def __init__(self, raw_lines: list[str]):
        super().__init__(raw_lines)
        self.name = raw_lines[0][len(self.GROUP_PREFIX) :]
        for i, line in enumerate(raw_lines):
            if line.startswith(self.GROUP_COLUMN_PREFIX):
                self.group_data = raw_lines[1 : i - 1]
        # TODO handle group data


# TODO do we need to do anything with these?
class NoteBlock(Block):
    BLOCK_TYPE = "Note"


def split_wavelengths(
    values: Optional[str], delimiter: str = " "
) -> Optional[list[int]]:
    if values is None:
        return None
    return [int(value) for value in values.split(delimiter)]


def get_well_coordinates(row_number: int, column_number: str) -> str:
    row_letters = ""
    while row_number > 0:
        row_number, remainder = divmod(row_number - 1, 26)
        row_letters = chr(START_LETTER_CODE + remainder) + row_letters
    return row_letters + str(column_number)


class WellData:
    values: list[Optional[float]]
    dimensions: list[Optional[PrimitiveValue]]
    wavelengths: list[Optional[int]]
    temperature: Optional[str]
    processed_data: list[float]

    def __init__(self) -> None:
        self.values = []
        self.dimensions = []
        self.wavelengths = []
        self.temperature = None
        self.processed_data = []

    def add_value(
        self,
        value: float,
        dimension: Optional[PrimitiveValue],
        temperature: Optional[str],
        wavelength: Optional[int],
    ) -> None:
        self.values.append(value)
        self.dimensions.append(dimension)
        self.wavelengths.append(wavelength)
        if temperature is not None:
            if self.temperature is not None and temperature != self.temperature:
                error = "Expected all measurements to have the same temperature"
                raise AllotropeConversionError(error)
            self.temperature = temperature

    @property
    def is_empty(self) -> bool:
        return not self.dimensions or not self.values


class PlateBlock(Block, ABC):
    BLOCK_TYPE = "Plate"
    CONCEPT: ClassVar[str]
    MANIFEST: ClassVar[str]
    READ_MODE: ClassVar[str]
    UNIT: ClassVar[str]
    DATA_TYPE_IDX: ClassVar[int]
    data_header: list[Optional[str]]
    data_type: Optional[str]
    export_format: Optional[str]
    name: Optional[str]
    kinetic_points: int
    num_columns: int
    num_rows: int
    num_wavelengths: Optional[int]
    num_wells: int
    pmt_gain: Optional[str]
    read_type: Optional[str]
    wavelengths: list[int]
    well_data: defaultdict[str, WellData]

    def __init__(
        self,
        header: list[Optional[str]],
        split_lines: list[list[Optional[str]]],
        raw_lines: list[str],
    ):
        super().__init__(raw_lines)
        self.parse_header(header)
        self.well_data = defaultdict(WellData)
        self.data_header = split_lines[1]
        data_lines = split_lines[2:]
        if self.export_format == ExportFormat.TIME_FORMAT.value:
            self._parse_time_format_data(data_lines)
        elif self.export_format == ExportFormat.PLATE_FORMAT.value:
            self._parse_plate_format_data(data_lines)
        else:
            error = f"unrecognized export format {self.export_format}"
            raise AllotropeConversionError(error)

    def _parse_reduced_plate_rows(
        self, reduced_data_rows: list[list[Optional[str]]]
    ) -> None:
        for i, row in enumerate(reduced_data_rows):
            for j, value in enumerate(row[2 : self.num_columns + 2]):
                if value is None:
                    continue
                col_number = assert_not_none(self.data_header[j + 2], "column number")
                well = get_well_coordinates(i + 1, col_number)
                self.well_data[well].processed_data.append(float(value))

    def _parse_reduced_columns(self, reduced_data_row: list[Optional[str]]) -> None:
        for i, value in enumerate(reduced_data_row[2:]):
            if value is None:
                continue
            well = assert_not_none(self.data_header[i + 2], "well")
            self.well_data[well].processed_data.append(float(value))

    def _add_data_point(
        self,
        well: str,
        value: str,
        data_key: Optional[str],
        temperature: Optional[str],
        wavelength: Optional[int],
    ) -> None:
        dimension = (
            wavelength if self.read_type == ReadType.ENDPOINT.value else data_key
        )
        self.well_data[well].add_value(
            value=float(value),
            dimension=dimension,
            temperature=temperature,
            wavelength=wavelength,
        )

    def _parse_time_format_data(self, data_lines: list[list[Optional[str]]]) -> None:
        if self.data_type == DataType.RAW.value:
            num_row_blocks = (
                self.num_wavelengths if self.num_wavelengths is not None else 1
            )
            for wavelength_index in range(num_row_blocks):
                start_index = wavelength_index * (self.kinetic_points + 1)
                wavelength_rows = data_lines[
                    start_index : start_index + self.kinetic_points
                ]
                for row in wavelength_rows:
                    for i, value in enumerate(row[2 : self.num_wells + 2]):
                        if value is None:
                            continue
                        well = assert_not_none(self.data_header[i + 2], "well")
                        self._add_data_point(
                            well,
                            value,
                            data_key=row[0],
                            temperature=row[1],
                            wavelength=self.get_wavelength(wavelength_index),
                        )
            if len(data_lines) > (self.kinetic_points + 1) * num_row_blocks:
                reduced_row = data_lines[-1][: self.num_wells + 2]
                self._parse_reduced_columns(reduced_row)
        elif self.data_type == DataType.REDUCED.value:
            reduced_row = data_lines[-1][: self.num_wells + 2]
            self._parse_reduced_columns(reduced_row)
        else:
            error = f"unrecognized data type {self.data_type}"
            raise AllotropeConversionError(error)

    def _parse_plate_format_data(self, data_lines: list[list[Optional[str]]]) -> None:
        end_raw_data_index = 0
        if self.data_type == DataType.RAW.value:
            for read_index in range(self.kinetic_points):
                start_index = read_index * (self.num_rows + 1)
                read_rows = data_lines[start_index : start_index + self.num_rows]
                data_key = read_rows[0][0]
                temperature = read_rows[0][1]
                for i, row in enumerate(read_rows):
                    wavelength_index = 0
                    num_row_blocks = (
                        self.num_wavelengths if self.num_wavelengths is not None else 1
                    )
                    for wavelength_index in range(num_row_blocks):
                        col_start_index = 2 + (
                            wavelength_index * (self.num_columns + 1)
                        )
                        for j, value in enumerate(
                            row[col_start_index : col_start_index + self.num_columns]
                        ):
                            if value is None:
                                continue
                            col_number = assert_not_none(
                                self.data_header[j + 2], "column number"
                            )
                            well = get_well_coordinates(i + 1, col_number)
                            self._add_data_point(
                                well,
                                value,
                                data_key=data_key,
                                temperature=temperature,
                                wavelength=self.get_wavelength(wavelength_index),
                            )
            end_raw_data_index = ((self.num_rows + 1) * self.kinetic_points) + 1
            reduced_data_rows = data_lines[end_raw_data_index:]
            if len(reduced_data_rows) == self.num_rows:
                self._parse_reduced_plate_rows(reduced_data_rows)
        elif self.data_type == DataType.REDUCED.value:
            reduced_data_rows = data_lines[end_raw_data_index:]
            self._parse_reduced_plate_rows(reduced_data_rows)
        else:
            error = f"unrecognized data type {self.data_type}"
            raise AllotropeConversionError(error)

    def parse_read_mode_header(self, header: list[Optional[str]]) -> None:
        raise NotImplementedError

    def parse_header(self, header: list[Optional[str]]) -> None:
        [
            _,  # Plate:
            self.name,
            self.export_version,
            self.export_format,
            self.read_type,
            _,  # Read mode
        ] = header[:6]
        if self.export_version != EXPORT_VERSION:
            error = f"Invalid export version {self.export_version}"
            raise AllotropeConversionError(error)

        [
            self.data_type,
            _,  # Pre-read, always FALSE
            kinetic_points,
            read_time_or_scan_pattern,
            read_interval_or_scan_density,
            self.start_wavelength,
            self.end_wavelength,
            self.wavelength_step,
            num_wavelengths,
            wavelengths_str,
            self.first_column,
            num_columns,
            num_wells,
        ] = header[self.DATA_TYPE_IDX : self.DATA_TYPE_IDX + 13]
        self.kinetic_points = assert_not_none(try_int(kinetic_points), "kinetic_points")
        self.num_columns = assert_not_none(try_int(num_columns), "num_columns")
        self.num_wells = assert_not_none(try_int(num_wells), "num_wells")
        self.num_wavelengths = try_int(num_wavelengths)
        self.wavelengths = split_wavelengths(wavelengths_str) or []
        self.parse_read_mode_header(header)

        if self.read_type == ReadType.KINETIC.value:
            self.read_time = read_time_or_scan_pattern
            self.read_interval = read_interval_or_scan_density
        elif self.read_type == ReadType.WELL_SCAN.value:
            self.scan_pattern = read_time_or_scan_pattern
            self.scan_density = read_interval_or_scan_density

    @property
    def is_single_wavelength(self) -> bool:
        return (
            self.read_type in {ReadType.SPECTRUM.value, ReadType.ENDPOINT.value}
            and len(self.wavelengths) == 1
        )

    @property
    def has_wavelength_dimension(self) -> bool:
        return (
            self.read_type not in {ReadType.SPECTRUM.value, ReadType.ENDPOINT.value}
            and len(self.wavelengths) > 1
        )

    def get_wavelength(self, wavelength_index: int) -> Optional[int]:
        if self.read_type == ReadType.SPECTRUM.value:
            return None
        else:
            return self.wavelengths[wavelength_index] if self.wavelengths else None

    def get_data_cube_dimensions(self) -> list[tuple[str, str, Optional[str]]]:
        dimensions: list[tuple[str, str, Optional[str]]] = []
        if self.read_type == ReadType.KINETIC.value:
            dimensions = [("double", "elapsed time", "s")]
        elif self.read_type == ReadType.WELL_SCAN.value:
            dimensions = [("int", "x", None)]
        elif self.read_type in {ReadType.SPECTRUM.value, ReadType.ENDPOINT.value}:
            dimensions = [("int", "wavelength", "nm")]
        else:
            error = f"cannot make data cube for {self.read_type}"
            raise AllotropeConversionError(error)
        if self.has_wavelength_dimension:
            dimensions.append(("int", "wavelength", "nm"))
        return dimensions

    def generate_sample_document(self, well: str) -> SampleDocument:
        return SampleDocument(well_location_identifier=well, plate_barcode=self.name)

    def generate_data_cube(self, well_data: WellData) -> TDatacube:
        dimension_data = [well_data.dimensions] + (
            [well_data.wavelengths] if self.has_wavelength_dimension else []
        )
        return TDatacube(
            cube_structure=TDatacubeStructure(
                [
                    TDatacubeComponent(FieldComponentDatatype(data_type), concept, unit)
                    for data_type, concept, unit in self.get_data_cube_dimensions()
                ],
                [
                    TDatacubeComponent(
                        FieldComponentDatatype("double"), self.CONCEPT, self.UNIT
                    )
                ],
            ),
            data=TDatacubeData(
                dimension_data,  # type: ignore[arg-type]
                [well_data.values],
            ),
        )

    def generate_processed_data_aggreate_document(
        self, well_data: WellData
    ) -> ProcessedDataAggregateDocument:
        return ProcessedDataAggregateDocument(
            [
                ProcessedDataDocumentItem(
                    val, data_processing_description="processed data"
                )
                for val in well_data.processed_data
            ]
        )

    def to_allotrope(self) -> Any:
        raise NotImplementedError



import pytest

from allotropy.parser_factory import Vendor
from allotropy.parsers.beckman_vi_cell_blu.vi_cell_blu_parser import ViCellBluParser
from allotropy.parsers.utils.timestamp_parser import TimestampParser
from tests.parsers.beckman_vi_cell_blu.vi_cell_blu_data import (
    get_data,
    get_filename,
    get_model,
)
from tests.parsers.test_utils import from_file, validate_contents, validate_schema

output_files = (
    "Beckman_Vi-Cell-BLU_example01",
    "Beckman_Vi-Cell-BLU_example02",
)

VENDOR_TYPE = Vendor.BECKMAN_VI_CELL_BLU
SCHEMA_FILE = "cell-counting/BENCHLING/2023/09/cell-counting.json"


@pytest.mark.parametrize("output_file", output_files)
def test_parse_vi_cell_blu_to_asm_schema_is_valid(output_file: str) -> None:
    test_filepath = f"tests/parsers/beckman_vi_cell_blu//testdata/{output_file}.csv"
    allotrope_dict = from_file(test_filepath, VENDOR_TYPE)
    validate_schema(allotrope_dict, SCHEMA_FILE)


@pytest.mark.parametrize("output_file", output_files)
def test_parse_vi_cell_blu_to_asm_expected_contents(output_file: str) -> None:
    test_filepath = f"tests/parsers/beckman_vi_cell_blu//testdata/{output_file}.csv"
    expected_filepath = f"tests/parsers/beckman_vi_cell_blu/testdata/{output_file}.json"
    allotrope_dict = from_file(test_filepath, VENDOR_TYPE)
    validate_contents(allotrope_dict, expected_filepath)


def test_get_model() -> None:
    parser = ViCellBluParser(TimestampParser())
    result = parser._get_model(get_data(), get_filename())

    cell_counting_document_item = (
        result.cell_counting_aggregate_document.cell_counting_document[0]
    )
    measurement_document_item = (
        cell_counting_document_item.measurement_aggregate_document.measurement_document[
            0
        ]
    )
    measurement_document_item.measurement_identifier = ""

    assert result == get_model()



from __future__ import annotations

from io import BytesIO
from typing import Optional

import pandas as pd
import pytest

from allotropy.allotrope.allotrope import AllotropeConversionError
from allotropy.allotrope.models.pcr_benchling_2023_09_qpcr import ExperimentType
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_builders import (
    DataBuilder,
    HeaderBuilder,
    ResultsBuilder,
)
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_structure import (
    Data,
    Header,
    Result,
    WellItem,
)
from allotropy.parsers.appbio_quantstudio.calculated_document import CalculatedDocument
from allotropy.parsers.appbio_quantstudio.referenceable import Referenceable
from allotropy.parsers.lines_reader import LinesReader
from tests.parsers.appbio_quantstudio.appbio_quantstudio_data import (
    get_broken_calc_doc_data,
    get_data,
    get_data2,
    get_genotyping_data,
    get_rel_std_curve_data,
)


def rm_uuid(data: Data) -> Data:
    for well in data.wells:
        if well.calculated_document:
            rm_uuid_calc_doc(well.calculated_document)

        for well_item in well.items.values():
            well_item.uuid = ""

    for calc_doc in data.calculated_documents:
        rm_uuid_calc_doc(calc_doc)

    return data


def rm_uuid_calc_doc(calc_doc: CalculatedDocument) -> None:
    calc_doc.uuid = ""
    for source in calc_doc.data_sources:
        if isinstance(source.reference, CalculatedDocument):
            rm_uuid_calc_doc(source.reference)
        elif isinstance(source.reference, Referenceable):
            source.reference.uuid = ""


@pytest.mark.short
def test_header_builder_returns_header_instance() -> None:
    header_contents = get_raw_header_contents()

    assert isinstance(HeaderBuilder.build(LinesReader(header_contents)), Header)


def test_header_builder() -> None:
    device_identifier = "device1"
    model_number = "123"
    device_serial_number = "1234"
    measurement_method_identifier = "measurement ID"
    pcr_detection_chemistry = "detection1"
    passive_reference_dye_setting = "blue"
    experimental_data_identifier = "data Identifier"

    header_contents = get_raw_header_contents(
        measurement_time="2010-10-01 01:44:54 AM EDT",
        plate_well_count="96 plates",
        experiment_type="Genotyping",
        device_identifier=device_identifier,
        model_number=model_number,
        device_serial_number=device_serial_number,
        measurement_method_identifier=measurement_method_identifier,
        pcr_detection_chemistry=pcr_detection_chemistry,
        passive_reference_dye_setting=passive_reference_dye_setting,
        experimental_data_identifier=experimental_data_identifier,
    )

    assert HeaderBuilder.build(LinesReader(header_contents)) == Header(
        measurement_time="2010-10-01 01:44:54 AM EDT",
        plate_well_count=96,
        experiment_type=ExperimentType.genotyping_qPCR_experiment,
        device_identifier=device_identifier,
        model_number=model_number,
        device_serial_number=device_serial_number,
        measurement_method_identifier=measurement_method_identifier,
        pcr_detection_chemistry=pcr_detection_chemistry,
        passive_reference_dye_setting=passive_reference_dye_setting,
        barcode=None,
        analyst=None,
        experimental_data_identifier=experimental_data_identifier,
    )


@pytest.mark.parametrize(
    "parameter,expected_error",
    [
        ("model_number", "Unable to get model number"),
        (
            "measurement_method_identifier",
            "Unable to get measurement method identifier",
        ),
        ("pcr_detection_chemistry", "Unable to get PCR detection chemistry"),
        ("plate_well_count", "Unable to interpret plate well count"),
    ],
)
@pytest.mark.short
def test_header_builder_required_parameter_none_then_raise(
    parameter: str, expected_error: str
) -> None:
    header_contents = get_raw_header_contents(**{parameter: None})

    with pytest.raises(AllotropeConversionError, match=expected_error):
        HeaderBuilder.build(LinesReader(header_contents))


@pytest.mark.short
def test_header_builder_invalid_plate_well_count() -> None:
    header_contents = get_raw_header_contents(plate_well_count="0 plates")

    with pytest.raises(AllotropeConversionError):
        HeaderBuilder.build(LinesReader(header_contents))


@pytest.mark.short
def test_header_builder_no_header_then_raise() -> None:
    header_contents = get_raw_header_contents(raw_text="")

    with pytest.raises(AllotropeConversionError):
        HeaderBuilder.build(LinesReader(header_contents))


@pytest.mark.short
def test_results_builder() -> None:

    data = pd.DataFrame(
        {
            "Well": [1],
            "SNP Assay Name": ["CYP19_2"],
            "Allele1 Automatic Ct Threshold": [True],
            "Allele1 Ct Threshold": [0.219],
            "Allele2 Ct Threshold": [9999],
            "Allele1 Ct": ["Undetermined"],
        }
    )
    well_item = WellItem(
        uuid="be566c58-41f3-40f8-900b-ef1dff20d264",
        identifier=1,
        position="A1",
        target_dna_description="CYP19_2-Allele 1",
        sample_identifier="NTC",
        well_location_identifier="A1",
        reporter_dye_setting="SYBR",
        quencher_dye_setting=None,
        sample_role_type="PC_ALLELE_1",
    )
    result = ResultsBuilder.build(
        data, well_item, ExperimentType.genotyping_qPCR_experiment
    )
    assert isinstance(result, Result)
    assert result.cycle_threshold_value_setting == 0.219
    assert result.cycle_threshold_result is None
    assert result.automatic_cycle_threshold_enabled_setting is True


@pytest.mark.short
@pytest.mark.parametrize(
    "test_filepath,expected_data",
    [
        (
            "tests/parsers/appbio_quantstudio/testdata/appbio_quantstudio_test01.txt",
            get_data(),
        ),
        (
            "tests/parsers/appbio_quantstudio/testdata/appbio_quantstudio_test02.txt",
            get_data2(),
        ),
        (
            "tests/parsers/appbio_quantstudio/testdata/appbio_quantstudio_test03.txt",
            get_genotyping_data(),
        ),
        (
            "tests/parsers/appbio_quantstudio/testdata/appbio_quantstudio_test04.txt",
            get_rel_std_curve_data(),
        ),
        (
            "tests/parsers/appbio_quantstudio/testdata/appbio_quantstudio_test05.txt",
            get_broken_calc_doc_data(),
        ),
    ],
)
def test_data_builder(test_filepath: str, expected_data: Data) -> None:
    with open(test_filepath, "rb") as raw_contents:
        reader = LinesReader(raw_contents)
    assert rm_uuid(DataBuilder.build(reader)) == rm_uuid(expected_data)


def get_raw_header_contents(
    raw_text: Optional[str] = None,
    measurement_time: Optional[str] = "2010-10-01 01:44:54 AM EDT",
    plate_well_count: Optional[str] = "96-Well Block (0.2mL)",
    experiment_type: Optional[str] = "Presence/Absence",
    device_identifier: Optional[str] = "278880034",
    model_number: Optional[str] = "QuantStudio(TM) 6 Flex System",
    device_serial_number: Optional[str] = "278880034",
    measurement_method_identifier: Optional[str] = "Ct",
    pcr_detection_chemistry: Optional[str] = "TAQMAN",
    passive_reference_dye_setting: Optional[str] = "ROX",
    barcode: Optional[str] = "NA",
    analyst: Optional[str] = "NA",
    experimental_data_identifier: Optional[
        str
    ] = "QuantStudio 96-Well Presence-Absence Example",
) -> BytesIO:
    if raw_text is not None:
        return BytesIO(raw_text.encode("utf-8"))

    header_dict = {
        "Experiment Run End Time": measurement_time,
        "Block Type": plate_well_count,
        "Experiment Type ": experiment_type,
        "Instrument Name": device_identifier,
        "Instrument Type": model_number,
        "Instrument Serial Number": device_serial_number,
        "Quantification Cycle Method": measurement_method_identifier,
        "Chemistry": pcr_detection_chemistry,
        "Passive Reference": passive_reference_dye_setting,
        "Experiment Barcode": barcode,
        "Experiment User Name": analyst,
        "Experiment Name": experimental_data_identifier,
    }

    raw_text = "\n".join(
        [
            f"* {header_name} = {header_value}"
            for header_name, header_value in header_dict.items()
            if header_value is not None
        ]
    )

    return BytesIO(raw_text.encode("utf-8"))



import io
from typing import Any

from allotropy.allotrope.allotrope import AllotropeConversionError
from allotropy.parsers.moldev_softmax_pro.block_factory import create_block
from allotropy.parsers.moldev_softmax_pro.plate_block import PlateBlock
from allotropy.parsers.vendor_parser import VendorParser

BLOCKS_LINE_PREFIX = "##BLOCKS= "
END_LINE_PREFIX = "~End"


class SoftmaxproParser(VendorParser):
    def _parse(self, contents: io.IOBase, filename: str) -> Any:  # noqa: ARG002
        file_contents = self._read_contents(contents)
        lines = file_contents.split("\n")

        # TODO parse whole file, not just blocks
        blocks_line = lines[0]
        if not blocks_line.startswith(BLOCKS_LINE_PREFIX):
            error = f"unrecognized start line {blocks_line}"
            raise AllotropeConversionError(error)

        num_blocks = int(blocks_line[len(BLOCKS_LINE_PREFIX) :])
        blocks = []
        current_block: list[str] = []
        for line in lines[1:]:
            if line.startswith(END_LINE_PREFIX):
                blocks.append(create_block(current_block))
                current_block = []
            elif current_block or line.strip():  # drop empty lines between blocks
                current_block.append(line)

        if len(blocks) != num_blocks:
            error = f"expected number of blocks to match {num_blocks}"
            raise AllotropeConversionError(error)

        plate_blocks = [block for block in blocks if isinstance(block, PlateBlock)]
        if len(plate_blocks) != 1:
            block_types = [block.BLOCK_TYPE for block in blocks]
            block_counts = {bt: block_types.count(bt) for bt in set(block_types)}
            error = f"expected exactly 1 plate block, got {block_counts}"
            raise AllotropeConversionError(error)
        return plate_blocks[0].to_allotrope()



from io import StringIO

from allotropy.allotrope.models.cell_culture_analyzer_benchling_2023_09_cell_culture_analyzer import (
    AnalyteAggregateDocument,
    AnalyteDocumentItem,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TNullableQuantityValueGramPerLiter,
    TNullableQuantityValueMillimeterOfMercury,
    TNullableQuantityValueMillimolePerLiter,
    TNullableQuantityValuePercent,
    TNullableQuantityValuePH,
)
from allotropy.parsers.novabio_flex2.constants import PROPERTY_MAPPINGS
from allotropy.parsers.novabio_flex2.novabio_flex2_structure import (
    Analyte,
    Data,
    Sample,
    SampleList,
    Title,
)


def get_input_title() -> str:
    return "SampleResults2022-06-28_142558.csv"


def get_input_stream() -> StringIO:
    data = {
        "Date & Time": "6-24-2022  14:34:52",
        "Sample ID": "BP_R10_KP_008_D0",
        "Sample Type": "Spent Media",
        "Gln": "1.83",
        "Glu": "0.33",
        "Gluc": "2.65",
        "Lac": "0.18",
        "NH4+": "0.48",
        "Na+": "151.6",
        "K+": "4.22",
        "Ca++": "0.82",
        "pH": "7.400",
        "PO2": "191.6",
        "PCO2": "46.8",
        "Vessel ID": "3L Bioreactor",
        "Batch ID": "KP_008",
        "Cell Type": "Workhog",
        "Vessel Temperature (°C)": "37.0",
        "Vessel Pressure (psi)": "30.0",
        "Sparging O2%": "320.9",
        "Chemistry Dilution Ratio": "1:1",
        "pH @ Temp": "7.400",
        "PCO2 @ Temp": "46.8",
        "PO2 @ Temp": "191.6",
        "O2 Saturation": "100.0",
        "CO2 Saturation": "6.6",
        "HCO3": "29.3",
        "Chemistry Flow Time": "6.67",
        "pH / Gas Flow Time": "3.65",
        "Tray Location": "",
        "Time In Tray": "",
        "Sample Time": "24-6-2022  14:34:52",
        "Operator": "BioprocessingTeam",
        "": "",
    }
    return StringIO("\n".join([",".join(data.keys()), ",".join(data.values())]))


def get_data() -> Data:
    return Data(
        title=Title(
            processing_time="2022-06-28 142558",
            device_identifier=None,
        ),
        sample_list=SampleList(
            analyst="BioprocessingTeam",
            samples=[
                Sample(
                    identifier="BP_R10_KP_008_D0",
                    role_type="Spent Media",
                    measurement_time="2022-06-24T14:34:52",
                    batch_identifier="KP_008",
                    analytes=sorted(
                        [
                            Analyte.create("Gln", 1.83),
                            Analyte.create("Glu", 0.33),
                            Analyte.create("Gluc", 2.65),
                            Analyte.create("Lac", 0.18),
                            Analyte.create("NH4+", 0.48),
                            Analyte.create("Na+", 151.6),
                            Analyte.create("K+", 4.22),
                            Analyte.create("Ca++", 0.82),
                            Analyte.create("HCO3", 29.3),
                        ]
                    ),
                    properties={
                        "co2_saturation": PROPERTY_MAPPINGS["co2_saturation"]["cls"](
                            value=6.6
                        ),
                        "o2_saturation": PROPERTY_MAPPINGS["o2_saturation"]["cls"](
                            value=100.0
                        ),
                        "pH": PROPERTY_MAPPINGS["pH"]["cls"](value=7.4),
                        "po2": PROPERTY_MAPPINGS["po2"]["cls"](value=191.6),
                        "pco2": PROPERTY_MAPPINGS["pco2"]["cls"](value=46.8),
                    },
                )
            ],
        ),
    )


def get_model() -> Model:
    return Model(
        measurement_aggregate_document=MeasurementAggregateDocument(
            measurement_identifier="",
            analyst="BioprocessingTeam",
            device_system_document=DeviceSystemDocument(
                device_identifier=None,
                model_number="NovaBio Flex2",
                device_serial_number=None,
            ),
            measurement_document=[
                MeasurementDocumentItem(
                    sample_document=SampleDocument(
                        sample_identifier="BP_R10_KP_008_D0",
                        batch_identifier="KP_008",
                        sample_role_type="Spent Media",
                    ),
                    measurement_time="2022-06-24T14:34:52+00:00",
                    analyte_aggregate_document=AnalyteAggregateDocument(
                        analyte_document=[
                            AnalyteDocumentItem(
                                analyte_name="ammonium",
                                molar_concentration=TNullableQuantityValueMillimolePerLiter(
                                    value=0.48,
                                    has_statistic_datum_role=None,
                                    unit="mmol/L",
                                    field_type=None,
                                ),
                            ),
                            AnalyteDocumentItem(
                                analyte_name="bicarbonate",
                                molar_concentration=TNullableQuantityValueMillimolePerLiter(
                                    value=29.3,
                                    has_statistic_datum_role=None,
                                    unit="mmol/L",
                                    field_type=None,
                                ),
                            ),
                            AnalyteDocumentItem(
                                analyte_name="calcium",
                                molar_concentration=TNullableQuantityValueMillimolePerLiter(
                                    value=0.82,
                                    has_statistic_datum_role=None,
                                    unit="mmol/L",
                                    field_type=None,
                                ),
                            ),
                            AnalyteDocumentItem(
                                analyte_name="glucose",
                                molar_concentration=TNullableQuantityValueGramPerLiter(
                                    value=2.65,
                                    has_statistic_datum_role=None,
                                    unit="g/L",
                                    field_type=None,
                                ),
                            ),
                            AnalyteDocumentItem(
                                analyte_name="glutamate",
                                molar_concentration=TNullableQuantityValueMillimolePerLiter(
                                    value=0.33,
                                    has_statistic_datum_role=None,
                                    unit="mmol/L",
                                    field_type=None,
                                ),
                            ),
                            AnalyteDocumentItem(
                                analyte_name="glutamine",
                                molar_concentration=TNullableQuantityValueMillimolePerLiter(
                                    value=1.83,
                                    has_statistic_datum_role=None,
                                    unit="mmol/L",
                                    field_type=None,
                                ),
                            ),
                            AnalyteDocumentItem(
                                analyte_name="lactate",
                                molar_concentration=TNullableQuantityValueGramPerLiter(
                                    value=0.18,
                                    has_statistic_datum_role=None,
                                    unit="g/L",
                                    field_type=None,
                                ),
                            ),
                            AnalyteDocumentItem(
                                analyte_name="potassium",
                                molar_concentration=TNullableQuantityValueMillimolePerLiter(
                                    value=4.22,
                                    has_statistic_datum_role=None,
                                    unit="mmol/L",
                                    field_type=None,
                                ),
                            ),
                            AnalyteDocumentItem(
                                analyte_name="sodium",
                                molar_concentration=TNullableQuantityValueMillimolePerLiter(
                                    value=151.6,
                                    has_statistic_datum_role=None,
                                    unit="mmol/L",
                                    field_type=None,
                                ),
                            ),
                        ]
                    ),
                    pco2=TNullableQuantityValueMillimeterOfMercury(
                        value=46.8,
                        has_statistic_datum_role=None,
                        unit="mmHg",
                        field_type=None,
                    ),
                    co2_saturation=TNullableQuantityValuePercent(
                        value=6.6,
                        has_statistic_datum_role=None,
                        unit="%",
                        field_type=None,
                    ),
                    po2=TNullableQuantityValueMillimeterOfMercury(
                        value=191.6,
                        has_statistic_datum_role=None,
                        unit="mmHg",
                        field_type=None,
                    ),
                    o2_saturation=TNullableQuantityValuePercent(
                        value=100.0,
                        has_statistic_datum_role=None,
                        unit="%",
                        field_type=None,
                    ),
                    optical_density=None,
                    pH=TNullableQuantityValuePH(
                        value=7.4,
                        has_statistic_datum_role=None,
                        unit="pH",
                        field_type=None,
                    ),
                    osmolality=None,
                    viability__cell_counter_=None,
                    total_cell_density__cell_counter_=None,
                    viable_cell_density__cell_counter_=None,
                    average_live_cell_diameter__cell_counter_=None,
                    average_total_cell_diameter__cell_counter_=None,
                    total_cell_diameter_distribution__cell_counter_=None,
                    viable_cell_count__cell_counter_=None,
                    total_cell_count__cell_counter_=None,
                )
            ],
            data_processing_time="2022-06-28T14:25:58+00:00",
        ),
        manifest="http://purl.allotrope.org/manifests/cell-culture-analyzer/BENCHLING/2023/09/cell-culture-analyzer.manifest",
    )



import pytest

from allotropy.allotrope.models.pcr_benchling_2023_09_qpcr import Model
from allotropy.parser_factory import Vendor
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_parser import (
    AppBioQuantStudioParser,
)
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_structure import Data
from allotropy.parsers.utils.timestamp_parser import TimestampParser
from tests.parsers.appbio_quantstudio.appbio_quantstudio_data import (
    get_broken_calc_doc_data,
    get_broken_calc_doc_model,
    get_data,
    get_data2,
    get_genotyping_data,
    get_genotyping_model,
    get_model,
    get_model2,
    get_rel_std_curve_data,
    get_rel_std_curve_model,
)
from tests.parsers.test_utils import from_file, validate_contents, validate_schema

output_files = (
    "appbio_quantstudio_example01",
    "appbio_quantstudio_example02",
    "appbio_quantstudio_example03",
    "appbio_quantstudio_example04",
    "appbio_quantstudio_example05",
    "appbio_quantstudio_example06",
    "appbio_quantstudio_example07",
    "appbio_quantstudio_example08",
)

VENDOR_TYPE = Vendor.APPBIO_QUANTSTUDIO


@pytest.mark.parametrize("output_file", output_files)
def test_parse_appbio_quantstudio_to_asm_schema(output_file: str) -> None:
    test_filepath = f"tests/parsers/appbio_quantstudio/testdata/{output_file}.txt"
    allotrope_dict = from_file(test_filepath, VENDOR_TYPE)
    validate_schema(allotrope_dict, "pcr/BENCHLING/2023/09/qpcr.json")


@pytest.mark.parametrize("output_file", output_files)
def test_parse_appbio_quantstudio_to_asm_contents(output_file: str) -> None:
    test_filepath = f"tests/parsers/appbio_quantstudio/testdata/{output_file}.txt"
    expected_filepath = test_filepath.replace(".txt", ".json")
    allotrope_dict = from_file(test_filepath, VENDOR_TYPE)

    validate_contents(allotrope_dict, expected_filepath)


@pytest.mark.short
@pytest.mark.parametrize(
    "file_name,data,model",
    [
        ("appbio_quantstudio_test01.txt", get_data(), get_model()),
        ("appbio_quantstudio_test02.txt", get_data2(), get_model2()),
        (
            "appbio_quantstudio_test03.txt",
            get_genotyping_data(),
            get_genotyping_model(),
        ),
        (
            "appbio_quantstudio_test04.txt",
            get_rel_std_curve_data(),
            get_rel_std_curve_model(),
        ),
        # test 5 will check the calculated data document structure when an
        # expected value is missing. In this case a path is borken and should
        # be entirely removed.
        (
            "appbio_quantstudio_test05.txt",
            get_broken_calc_doc_data(),
            get_broken_calc_doc_model(),
        ),
    ],
)
def test_get_model(file_name: str, data: Data, model: Model) -> None:
    parser = AppBioQuantStudioParser(TimestampParser())
    generated = parser._get_model(data, file_name)
    assert generated == model



import json
from typing import Any

from deepdiff import DeepDiff
import jsonschema
import numpy as np
import pandas as pd

from allotropy.allotrope.schemas import get_schema
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.parser_factory import VendorType
from allotropy.to_allotrope import allotrope_from_file


def replace_asm_converter_name_and_version(allotrope_dict: dict) -> None:
    for key, value in allotrope_dict.items():
        if key == "data system document":
            value["ASM converter name"] = ASM_CONVERTER_NAME
            value["ASM converter version"] = ASM_CONVERTER_VERSION
        if isinstance(value, dict):
            replace_asm_converter_name_and_version(value)


def assert_allotrope_dicts_equal(expected: dict, actual: dict) -> None:
    replace_asm_converter_name_and_version(expected)
    exclude_regex = [
        r"\['measurement identifier'\]",
        r"\['data source identifier'\]",
        r"\['calculated data identifier'\]",
    ]
    # Uncomment for more info
    # print(DeepDiff(expected, actual, exclude_regex_paths=exclude_regex))
    assert not DeepDiff(
        expected,
        actual,
        exclude_regex_paths=exclude_regex,
        ignore_type_in_groups=[(float, np.float64)],
    )


def from_file(test_file: str, vendor_type: VendorType) -> dict[str, Any]:
    return allotrope_from_file(test_file, vendor_type)


def validate_schema(allotrope_dict: dict[str, Any], schema_relative_path: str) -> None:
    allotrope_schema = get_schema(schema_relative_path)
    jsonschema.validate(
        allotrope_dict,
        allotrope_schema,
        format_checker=jsonschema.validators.Draft202012Validator.FORMAT_CHECKER,
    )


def validate_contents(allotrope_dict: dict[str, Any], expected_file: str) -> None:
    with open(expected_file) as f:
        expected_dict = json.load(f)
        assert_allotrope_dicts_equal(expected_dict, allotrope_dict)


def build_series(elements: list[tuple]) -> pd.Series:
    index, data = list(zip(*elements))
    # pd.Series has a Generic.
    return pd.Series(data=data, index=index)  # type: ignore[no-any-return]



import io
from typing import Any
import uuid

import pandas as pd

from allotropy.allotrope.models.cell_counting_benchling_2023_09_cell_counting import (
    CellCountingAggregateDocument,
    CellCountingDocumentItem,
    DataProcessingDocument,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    ProcessedDataDocument,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCell,
    TQuantityValueMicrometer,
    TQuantityValueMillionCellsPerMilliliter,
    TQuantityValuePercent,
    TQuantityValueUnitless,
)
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.parsers.beckman_vi_cell_blu.constants import (
    DEFAULT_ANALYST,
    DEFAULT_MODEL_NUMBER,
    VICELL_BLU_SOFTWARE_NAME,
)
from allotropy.parsers.beckman_vi_cell_blu.vi_cell_blu_reader import ViCellBluReader
from allotropy.parsers.vendor_parser import VendorParser

property_lookup = {
    "Average viable diameter (μm)": TQuantityValueMicrometer,
    "Average circularity": TQuantityValueUnitless,
    "Average diameter (μm)": TQuantityValueMicrometer,
    "Average viable circularity": TQuantityValueUnitless,
    "Dilution": TQuantityValueUnitless,
    "Maximum Diameter (μm)": TQuantityValueMicrometer,
    "Minimum Diameter (μm)": TQuantityValueMicrometer,
    "Cell count": TQuantityValueCell,
    "Total (x10^6) cells/mL": TQuantityValueMillionCellsPerMilliliter,
    "Viability (%)": TQuantityValuePercent,
    "Viable cells": TQuantityValueCell,
    "Viable (x10^6) cells/mL": TQuantityValueMillionCellsPerMilliliter,
}


def get_property_from_sample(sample: pd.Series, property_name: str) -> Any:
    return property_lookup[property_name](value=value) if (value := sample.get(property_name)) else None  # type: ignore[arg-type]


class ViCellBluParser(VendorParser):
    def _parse(self, contents: io.IOBase, filename: str) -> Model:
        return self._get_model(ViCellBluReader.read(contents), filename)

    def _get_model(self, data: pd.DataFrame, filename: str) -> Model:
        return Model(
            cell_counting_aggregate_document=CellCountingAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    model_number=DEFAULT_MODEL_NUMBER,
                ),
                data_system_document=DataSystemDocument(
                    file_name=filename,
                    software_name=VICELL_BLU_SOFTWARE_NAME,
                    ASM_converter_name=ASM_CONVERTER_NAME,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                cell_counting_document=self._get_cell_counting_document(data),
            ),
        )

    def _get_cell_counting_document(
        self, data: pd.DataFrame
    ) -> list[CellCountingDocumentItem]:
        return [
            self._get_cell_counting_document_item(sample)
            for _, sample in data.iterrows()
            if sample.get("Cell count")
        ]

    def _get_cell_counting_document_item(
        self, sample: pd.Series
    ) -> CellCountingDocumentItem:
        return CellCountingDocumentItem(
            analyst=sample.get("Analysis by") or DEFAULT_ANALYST,  # type: ignore[arg-type]
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_document=[
                    MeasurementDocumentItem(
                        measurement_identifier=str(uuid.uuid4()),
                        measurement_time=self.get_date_time(
                            sample.get("Analysis date/time")
                        ),
                        sample_document=SampleDocument(sample_identifier=sample.get("Sample ID")),  # type: ignore[arg-type]
                        device_control_aggregate_document=DeviceControlAggregateDocument(
                            device_control_document=[
                                DeviceControlDocumentItem(
                                    device_type="brightfield imager (cell counter)",
                                    detection_type="brightfield",
                                )
                            ]
                        ),
                        processed_data_document=ProcessedDataDocument(
                            data_processing_document=DataProcessingDocument(
                                cell_type_processing_method=sample.get("Cell type"),  # type: ignore[arg-type]
                                minimum_cell_diameter=get_property_from_sample(
                                    sample, "Minimum Diameter (μm)"
                                ),
                                maximum_cell_diameter=get_property_from_sample(
                                    sample, "Maximum Diameter (μm)"
                                ),
                                cell_density_dilution_factor=get_property_from_sample(
                                    sample, "Dilution"
                                ),
                            ),
                            viability__cell_counter_=get_property_from_sample(
                                sample, "Viability (%)"
                            ),
                            viable_cell_density__cell_counter_=get_property_from_sample(
                                sample, "Viable (x10^6) cells/mL"
                            ),
                            total_cell_count=get_property_from_sample(
                                sample, "Cell count"
                            ),
                            total_cell_density__cell_counter_=get_property_from_sample(
                                sample, "Total (x10^6) cells/mL"
                            ),
                            average_total_cell_diameter=get_property_from_sample(
                                sample, "Average diameter (μm)"
                            ),
                            average_live_cell_diameter__cell_counter_=get_property_from_sample(
                                sample, "Average viable diameter (μm)"
                            ),
                            viable_cell_count=get_property_from_sample(
                                sample, "Viable cells"
                            ),
                            average_total_cell_circularity=get_property_from_sample(
                                sample, "Average circularity"
                            ),
                            average_viable_cell_circularity=get_property_from_sample(
                                sample, "Average viable circularity"
                            ),
                        ),
                    )
                ],
            ),
        )



import pytest

from allotropy.allotrope.allotrope import AllotropeConversionError
from allotropy.parser_factory import Vendor
from tests.parsers.test_utils import from_file, validate_contents, validate_schema

VENDOR_TYPE = Vendor.MOLDEV_SOFTMAX_PRO


@pytest.mark.parametrize(
    "test_file,schema_file,expected_file",
    [
        (
            "tests/parsers/moldev_softmax_pro/testdata/fl_kinetic_plates.txt",
            "fluorescence/BENCHLING/2023/09/fluorescence.json",
            "tests/parsers/moldev_softmax_pro/testdata/expected_fl_kinetic.json",
        ),
        (
            "tests/parsers/moldev_softmax_pro/testdata/lum_spectrum_columns.txt",
            "luminescence/BENCHLING/2023/09/luminescence.json",
            "tests/parsers/moldev_softmax_pro/testdata/expected_lum_spectrum.json",
        ),
        (
            "tests/parsers/moldev_softmax_pro/testdata/abs_endpoint_plates.txt",
            "ultraviolet-absorbance/BENCHLING/2023/09/ultraviolet-absorbance.json",
            "tests/parsers/moldev_softmax_pro/testdata/expected_abs_endpoint.json",
        ),
    ],
)
def test_to_allotrope(test_file: str, schema_file: str, expected_file: str) -> None:
    allotrope_dict = from_file(test_file, VENDOR_TYPE)
    validate_schema(allotrope_dict, schema_file)
    validate_contents(allotrope_dict, expected_file)
    assert (
        len(allotrope_dict["measurement aggregate document"]["measurement document"])
        == 96
    )


def test_handles_unrecognized_read_mode() -> None:
    with pytest.raises(
        AllotropeConversionError, match="unrecognized read mode Time Resolved"
    ):
        from_file(
            "tests/parsers/moldev_softmax_pro/testdata/trf_well_scan_plates.txt",
            VENDOR_TYPE,
        )



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Optional, Union

from allotropy.allotrope.models.shared.definitions.definitions import (
    TDatacube,
    TStringValue,
)


class SampleRoleType(Enum):
    control_sample_role = "control sample role"
    standard_sample_role = "standard sample role"
    validation_sample_role = "validation sample role"
    experiment_sample_role = "experiment sample role"
    sample_role = "sample role"
    spiked_sample_role = "spiked sample role"
    blank_role = "blank role"
    unknown_sample_role = "unknown sample role"
    undefined_sample_role = "undefined sample role"
    calibration_sample_role = "calibration sample role"
    unspiked_sample_role = "unspiked sample role"
    specimen_role = "specimen role"
    quality_control_sample_role = "quality control sample role"
    reference_sample_role = "reference sample role"


@dataclass
class SampleDocument:
    well_location_identifier: TStringValue
    sample_identifier: Optional[TStringValue] = None
    batch_identifier: Optional[TStringValue] = None
    sample_role_type: Optional[SampleRoleType] = None
    # TODO: plate_barcode isn't actually in the schema yet, needs to be modified to add it.
    plate_barcode: Optional[str] = None


@dataclass
class ProcessedDataDocumentItem:
    processed_data: Union[float, str, TDatacube]
    data_format_specification_type: Optional[TStringValue] = None
    data_processing_description: Optional[TStringValue] = None


@dataclass
class ProcessedDataAggregateDocument:
    processed_data_document: Optional[list[ProcessedDataDocumentItem]] = None



import io
from typing import Any, Optional
import uuid

import pandas as pd

from allotropy.allotrope.allotrope import AllotropeConversionError
from allotropy.allotrope.models.cell_counting_benchling_2023_09_cell_counting import (
    CellCountingAggregateDocument,
    CellCountingDocumentItem,
    DataProcessingDocument,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    ProcessedDataDocument,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCell,
    TQuantityValueMicrometer,
    TQuantityValueMillionCellsPerMilliliter,
    TQuantityValuePercent,
    TQuantityValueUnitless,
)
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.parsers.beckman_vi_cell_xr.constants import (
    DATE_HEADER,
    DEFAULT_ANALYST,
    MODEL_NUMBER,
    SOFTWARE_NAME,
    XrVersion,
)
from allotropy.parsers.beckman_vi_cell_xr.vi_cell_xr_reader import ViCellXRReader
from allotropy.parsers.vendor_parser import VendorParser

property_lookup = {
    "Dilution factor": TQuantityValueUnitless,
    "Total cells/ml (x10^6)": TQuantityValueMillionCellsPerMilliliter,
    "Avg. diam. (microns)": TQuantityValueMicrometer,
    "Viable cells": TQuantityValueCell,
    "Avg. circ.": TQuantityValueUnitless,
}


def get_property_from_sample(sample: pd.Series, property_name: str) -> Any:
    return property_lookup[property_name](value=value) if (value := sample.get(property_name)) else None  # type: ignore[arg-type]


class ViCellXRParser(VendorParser):
    def _parse(self, contents: io.IOBase, filename: str) -> Model:
        reader = ViCellXRReader(contents)

        return Model(
            cell_counting_aggregate_document=CellCountingAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    model_number=MODEL_NUMBER,
                    device_serial_number=self._get_device_serial_number(
                        reader.file_info
                    ),
                ),
                data_system_document=DataSystemDocument(
                    file_name=filename,
                    software_name=SOFTWARE_NAME,
                    software_version=reader.file_version.value,
                    ASM_converter_name=ASM_CONVERTER_NAME,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                cell_counting_document=[
                    self._get_cell_counting_document_item(sample, reader.file_version)
                    for _, sample in reader.data.iterrows()
                ],
            ),
        )

    def _get_device_serial_number(self, file_info: pd.Series) -> Optional[str]:
        serial = str(file_info["serial"])
        try:
            serial_number = serial[serial.rindex(":") + 1 :].strip()
        except ValueError:
            return None

        return serial_number

    def _get_cell_counting_document_item(
        self, sample: pd.Series, file_version: XrVersion
    ) -> CellCountingDocumentItem:
        # Required fields
        try:
            viability__cell_counter_ = TQuantityValuePercent(
                value=sample["Viability (%)"]
            )
            total_cell_count = TQuantityValueCell(value=sample["Total cells"])
            viable_cell_density__cell_counter_ = (
                TQuantityValueMillionCellsPerMilliliter(
                    value=sample["Viable cells/ml (x10^6)"]
                )
            )
        except KeyError as e:
            error = f"required value not found {e}"
            raise AllotropeConversionError(error) from e

        return CellCountingDocumentItem(
            analyst=DEFAULT_ANALYST,
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_document=[
                    MeasurementDocumentItem(
                        measurement_identifier=str(uuid.uuid4()),
                        measurement_time=self.get_date_time(
                            sample.get(DATE_HEADER[file_version])
                        ),
                        sample_document=SampleDocument(
                            sample_identifier=sample["Sample ID"]
                        ),
                        device_control_aggregate_document=DeviceControlAggregateDocument(
                            device_control_document=[
                                DeviceControlDocumentItem(
                                    device_type="brightfield imager (cell counter)",
                                    detection_type="brightfield",
                                )
                            ]
                        ),
                        processed_data_document=ProcessedDataDocument(
                            data_processing_document=DataProcessingDocument(
                                cell_type_processing_method=sample.get("Cell type"),  # type: ignore[arg-type]
                                cell_density_dilution_factor=get_property_from_sample(
                                    sample, "Dilution factor"
                                ),
                            ),
                            viability__cell_counter_=viability__cell_counter_,
                            viable_cell_density__cell_counter_=viable_cell_density__cell_counter_,
                            total_cell_count=total_cell_count,
                            total_cell_density__cell_counter_=get_property_from_sample(
                                sample, "Total cells/ml (x10^6)"
                            ),
                            average_total_cell_diameter=get_property_from_sample(
                                sample, "Avg. diam. (microns)"
                            ),
                            viable_cell_count=get_property_from_sample(
                                sample, "Viable cells"
                            ),
                            average_total_cell_circularity=get_property_from_sample(
                                sample, "Avg. circ."
                            ),
                        ),
                    )
                ],
            ),
        )



from io import StringIO

import pandas as pd

from allotropy.allotrope.models.cell_culture_analyzer_benchling_2023_09_cell_culture_analyzer import (
    AnalyteAggregateDocument,
    AnalyteDocumentItem,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TNullableQuantityValueMillimolePerLiter,
)
from allotropy.parsers.roche_cedex_bioht.roche_cedex_bioht_reader import (
    RocheCedexBiohtReader,
)
from allotropy.parsers.roche_cedex_bioht.roche_cedex_bioht_structure import (
    Analyte,
    AnalyteList,
    Data,
    Sample,
    Title,
)


def get_data_stream() -> StringIO:
    title = [
        "0",
        "2021-06-01 13:04:06",
        "#ARC-FILE#",
        "1.1a",
        "2021-05-01",
        "2021-06-01",
        "CEDEX BIO HT",
        "620103",
        "6.0.0.1905 (1905)",
        "ADMIN",
    ]

    body = [
        [
            "40",
            "2021-05-20 16:55:51",
            "PPDTEST1",
            "",
            "SAM",
            "        ",
            "GLN2B",
            "        ",
            "mmol/L",
            " ",
            "  2.45",
            "  0.17138",
            "R",
        ],
        [
            "40",
            "2021-05-20 16:55:51",
            "PPDTEST1",
            "",
            "Sample",
            "        ",
            "NH3LB",
            "        ",
            "mmol/L",
            " ",
            "  1.846",
            "0",
            "R",
        ],
    ]
    title_text = "\t".join(title)
    body_text = "\n".join(["\t".join(row) for row in body])
    return StringIO("\n".join([title_text, body_text]))


def get_reader() -> RocheCedexBiohtReader:
    return RocheCedexBiohtReader(get_data_stream())


def get_reader_title() -> pd.Series:
    return pd.Series(
        {
            "row type": 0,
            "data processing time": "2021-06-01 13:04:06",
            "col3": "#ARC-FILE#",
            "col4": "1.1a",
            "col5": "2021-05-01",
            "col6": "2021-06-01",
            "model number": "CEDEX BIO HT",
            "device serial number": 620103,
            "col9": "6.0.0.1905 (1905)",
            "analyst": "ADMIN",
        }
    )


def get_reader_samples() -> pd.DataFrame:
    return pd.DataFrame(
        columns=[
            "row type",
            "measurement time",
            "sample identifier",
            "batch identifier",
            "sample role type",
            "col6",
            "analyte name",
            "col8",
            "concentration unit",
            "flag",
            "concentration value",
            "col12",
            "col13",
        ],
        data=[
            [
                40,
                "2021-05-20 16:55:51",
                "PPDTEST1",
                "",
                "Sample",
                "        ",
                "glutamine",
                "        ",
                "mmol/L",
                " ",
                2.45,
                0.17138,
                "R",
            ],
            [
                40,
                "2021-05-20 16:55:51",
                "PPDTEST1",
                "",
                "Sample",
                "        ",
                "ammonia",
                "        ",
                "mmol/L",
                " ",
                1.846,
                0,
                "R",
            ],
        ],
    )


def get_data() -> Data:
    return Data(
        title=Title(
            data_processing_time="2021-06-01 13:04:06",
            analyst="ADMIN",
            model_number="CEDEX BIO HT",
            device_serial_number="620103",
        ),
        samples=[
            Sample(
                name="PPDTEST1",
                role_type="Sample",
                measurement_time="2021-05-20 16:55:51",
                analyte_list=AnalyteList(
                    analytes=[
                        Analyte("ammonia", 1.846, "mmol/L"),
                        Analyte("glutamine", 2.45, "mmol/L"),
                    ],
                    molar_concentration_dict={
                        "ammonia": [
                            TNullableQuantityValueMillimolePerLiter(
                                value=1.846,
                            )
                        ],
                        "glutamine": [
                            TNullableQuantityValueMillimolePerLiter(
                                value=2.45,
                            )
                        ],
                    },
                    molar_concentration_nans={},
                    non_aggregrable_dict={},
                    non_aggregable_nans={},
                    num_measurement_docs=1,
                ),
            )
        ],
    )


def get_model() -> Model:
    return Model(
        measurement_aggregate_document=MeasurementAggregateDocument(
            measurement_identifier="",
            analyst="ADMIN",
            device_system_document=DeviceSystemDocument(
                device_identifier=None,
                model_number="CEDEX BIO HT",
                device_serial_number="620103",
            ),
            measurement_document=[
                MeasurementDocumentItem(
                    sample_document=SampleDocument(
                        sample_identifier="PPDTEST1",
                        batch_identifier=None,
                        sample_role_type="Sample",
                    ),
                    measurement_time="2021-05-20T16:55:51+00:00",
                    analyte_aggregate_document=AnalyteAggregateDocument(
                        analyte_document=[
                            AnalyteDocumentItem(
                                analyte_name="ammonia",
                                molar_concentration=TNullableQuantityValueMillimolePerLiter(
                                    value=1.846,
                                ),
                            ),
                            AnalyteDocumentItem(
                                analyte_name="glutamine",
                                molar_concentration=TNullableQuantityValueMillimolePerLiter(
                                    value=2.45,
                                ),
                            ),
                        ]
                    ),
                    pco2=None,
                    co2_saturation=None,
                    po2=None,
                    o2_saturation=None,
                    optical_density=None,
                    pH=None,
                    osmolality=None,
                    viability__cell_counter_=None,
                    total_cell_density__cell_counter_=None,
                    viable_cell_density__cell_counter_=None,
                    average_live_cell_diameter__cell_counter_=None,
                    average_total_cell_diameter__cell_counter_=None,
                    total_cell_diameter_distribution__cell_counter_=None,
                    viable_cell_count__cell_counter_=None,
                    total_cell_count__cell_counter_=None,
                )
            ],
            data_processing_time="2021-06-01T13:04:06+00:00",
        ),
        manifest="http://purl.allotrope.org/manifests/cell-culture-analyzer/BENCHLING/2023/09/cell-culture-analyzer.manifest",
    )



from abc import ABC, abstractmethod
import io
from typing import Any, Optional

import chardet

from allotropy.allotrope.allotrope import AllotropeConversionError
from allotropy.allotrope.models.shared.definitions.definitions import TDateTimeValue
from allotropy.parsers.utils.timestamp_parser import TimestampParser
from allotropy.parsers.utils.values import assert_not_none


class VendorParser(ABC):
    def __init__(self, timestamp_parser: TimestampParser):
        self.timestamp_parser = timestamp_parser

    @abstractmethod
    def _parse(self, contents: io.IOBase, filename: str) -> Any:
        raise NotImplementedError

    def to_allotrope(self, contents: io.IOBase, filename: str) -> Any:
        return self._parse(contents, filename)

    def _read_contents(self, contents: io.IOBase) -> Any:
        file_bytes = contents.read()
        encoding = chardet.detect(file_bytes)["encoding"]
        if not encoding:
            error = "Did not detect encoding in input file"
            raise AllotropeConversionError(error)
        return file_bytes.decode(encoding)

    def parse_timestamp(self, time: Optional[str]) -> Optional[str]:
        return self.timestamp_parser.parse(time)

    def get_date_time(self, time: Any) -> TDateTimeValue:
        return assert_not_none(
            self.parse_timestamp(str(time)), msg=f"Invalid timestamp: '{time}'"
        )



import io
import uuid

from allotropy.allotrope.models.cell_culture_analyzer_benchling_2023_09_cell_culture_analyzer import (
    AnalyteAggregateDocument,
    AnalyteDocumentItem,
    DeviceSystemDocument,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
    SampleDocument,
)
from allotropy.parsers.roche_cedex_bioht.roche_cedex_bioht_reader import (
    RocheCedexBiohtReader,
)
from allotropy.parsers.roche_cedex_bioht.roche_cedex_bioht_structure import Data, Sample
from allotropy.parsers.vendor_parser import VendorParser


class RocheCedexBiohtParser(VendorParser):
    def _parse(self, contents: io.IOBase, filename: str) -> Model:  # noqa: ARG002
        reader = RocheCedexBiohtReader(contents)
        return self._get_model(Data.create(reader))

    def _get_model(self, data: Data) -> Model:
        return Model(
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_identifier=str(uuid.uuid4()),
                data_processing_time=self.get_date_time(
                    data.title.data_processing_time
                ),
                analyst=data.title.analyst,
                device_system_document=DeviceSystemDocument(
                    model_number=data.title.model_number,
                    device_serial_number=data.title.device_serial_number,
                ),
                measurement_document=self._get_measurement_document(data),
            )
        )

    def _get_measurement_document(self, data: Data) -> list[MeasurementDocumentItem]:
        measurement_document = []
        for sample in data.samples:
            measurement_document.extend(self._get_measurements_from_sample(sample))
        return measurement_document

    def _get_measurements_from_sample(
        self, sample: Sample
    ) -> list[MeasurementDocumentItem]:
        sample_measurements = [
            self._create_sample_measurement(sample)
            for _ in range(sample.analyte_list.num_measurement_docs)
        ]

        for (
            analyte_name,
            molar_concentrations,
        ) in sample.analyte_list.molar_concentration_dict.items():
            for sample_measurement, molar_concentration in zip(
                sample_measurements, molar_concentrations
            ):
                sample_measurement.analyte_aggregate_document.analyte_document.append(  # type: ignore[union-attr]
                    AnalyteDocumentItem(
                        analyte_name=analyte_name,
                        molar_concentration=molar_concentration,
                    )
                )

        for analyte_name, values in sample.analyte_list.non_aggregrable_dict.items():
            for sample_measurement, value in zip(sample_measurements, values):
                setattr(sample_measurement, analyte_name, value)

        return sample_measurements

    def _create_sample_measurement(self, sample: Sample) -> MeasurementDocumentItem:
        return MeasurementDocumentItem(
            sample_document=SampleDocument(
                sample_identifier=sample.name,
                sample_role_type=sample.role_type,
                batch_identifier=sample.batch,
            ),
            measurement_time=self.get_date_time(sample.measurement_time),
            analyte_aggregate_document=AnalyteAggregateDocument(analyte_document=[]),
        )



import uuid

from allotropy.allotrope.models.shared.definitions.custom import TQuantityValueNumber
from allotropy.allotrope.models.ultraviolet_absorbance_benchling_2023_09_ultraviolet_absorbance import (
    ContainerType,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    Model,
)
from allotropy.parsers.agilent_gen5.absorbance_data_point import AbsorbanceDataPoint
from allotropy.parsers.agilent_gen5.constants import ReadMode
from allotropy.parsers.agilent_gen5.plate_data import PlateData


class AbsorbancePlateData(PlateData):
    read_mode = ReadMode.ABSORBANCE
    data_point_cls = AbsorbanceDataPoint

    def to_allotrope(self, measurement_docs: list[MeasurementDocumentItem]) -> Model:
        return Model(
            measurement_aggregate_document=MeasurementAggregateDocument(
                measurement_identifier=str(uuid.uuid4()),
                measurement_time=self.datetime,
                analytical_method_identifier=self.protocol_file_path,
                experimental_data_identifier=self.experiment_file_path,
                container_type=ContainerType.well_plate,
                plate_well_count=TQuantityValueNumber(len(self.wells)),
                # TODO read_type=self.read_type.value?,
                measurement_document=measurement_docs,
            )
        )



from allotropy.allotrope.schema_parser.schema_model import (
    get_all_schema_components,
    snake_to_upper_camel,
)


def test_snake_to_upper_camel() -> None:
    assert "SampleDocument" == snake_to_upper_camel("sample document", " ")


def test_get_all_schema_components() -> None:
    schema = {
        "type": "object",
        "properties": {
            "prop1": {"type": "string"},
            "nested_prop": {
                "type": "object",
                "properties": {
                    "nested_prop2": {"type": "string"},
                },
            },
        },
    }

    assert {
        "prop1": {"type": "string"},
        "nested_prop": {
            "type": "object",
            "properties": {
                "nested_prop2": {"type": "string"},
            },
        },
        "nested_prop2": {"type": "string"},
    } == get_all_schema_components(schema)


def test_get_all_schema_components_list() -> None:
    schema = {
        "type": "array",
        "items": {"type": "object", "properties": {"prop1": {"type": "string"}}},
    }

    assert {"prop1": {"type": "string"}} == get_all_schema_components(schema)


def test_get_all_schema_components_all_of() -> None:
    schema = {
        "type": "array",
        "items": {
            "allOf": [
                {
                    "type": "object",
                    "properties": {
                        "allOf prop1": {
                            "type": "object",
                            "properties": {"value": {"type": "number"}},
                        },
                        "allOf prop2": {"type": "boolean"},
                    },
                },
                {
                    "properties": {
                        "allOf prop3": {"type": "string"},
                    }
                },
            ]
        },
        "minItems": 0,
    }

    assert {
        "allOf prop1": {
            "type": "object",
            "properties": {"value": {"type": "number"}},
        },
        "value": {"type": "number"},
        "allOf prop2": {"type": "boolean"},
        "allOf prop3": {"type": "string"},
    } == get_all_schema_components(schema)


def test_get_all_schema_components_one_of() -> None:
    schema = {
        "type": "object",
        "properties": {
            "oneOf prop": {
                "oneOf": [
                    {"type": "string"},
                    {
                        "type": "object",
                        "properties": {
                            "value": {"type": "string"},
                        },
                    },
                    {
                        "type": "object",
                        "properties": {
                            "other": {"type": "string"},
                        },
                    },
                ],
            },
        },
    }

    assert {
        "oneOf prop item": {
            "type": "object",
            "properties": {
                "value": {"type": "string"},
            },
        },
        "value": {"type": "string"},
        "oneOf prop item1": {
            "type": "object",
            "properties": {
                "other": {"type": "string"},
            },
        },
        "other": {"type": "string"},
    } == get_all_schema_components(schema)



from io import StringIO

import pytest

from allotropy.allotrope.models.fluorescence_benchling_2023_09_fluorescence import (
    ScanPositionSettingPlateReader,
)
from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.perkin_elmer_envision.perkin_elmer_envision_structure import (
    BasicAssayInfo,
    create_plate_maps,
    Filter,
    Instrument,
    Labels,
    Plate,
    PlateInfo,
    PlateMap,
    PlateType,
    Result,
)


def get_reader_from_lines(lines: list[str]) -> CsvReader:
    return CsvReader(StringIO("\n".join(lines)))


def test_create_plate_info() -> None:
    reader = get_reader_from_lines(
        [
            "Plate information",
            "Plate,Repeat,Barcode,Measured height,Chamber temperature at start,Measinfo,Kinetics,Measurement date,",
            '1,1,"=""BAR_123""",11.9,23.17,De=1st Ex=Top Em=Top Wdw=1 (14),0,10/13/2022 3:08:06 PM,',
        ]
    )

    expected = PlateInfo(
        number="1",
        barcode="BAR_123",
        emission_filter_id="1st",
        measurement_time="10/13/2022 3:08:06 PM",
        measured_height=11.9,
        chamber_temperature_at_start=23.17,
    )

    assert PlateInfo.create(reader) == expected


def test_create_plate_info_default_barcode() -> None:
    reader = get_reader_from_lines(
        [
            "Plate information",
            "Plate,Repeat,Barcode,Measured height,Chamber temperature at start,Measinfo,Kinetics,Measurement date,",
            '1,1,"=""""",11.9,23.17,De=1st Ex=Top Em=Top Wdw=1 (14),0,10/13/2022 3:08:06 PM,',
        ]
    )
    plate = PlateInfo.create(reader)
    assert plate and plate.barcode == "Plate 1"
    reader = get_reader_from_lines(
        [
            "Plate information",
            "Plate,Repeat,Barcode,Measured height,Chamber temperature at start,Measinfo,Kinetics,Measurement date,",
            "1,1,,11.9,23.17,De=1st Ex=Top Em=Top Wdw=1 (14),0,10/13/2022 3:08:06 PM,",
        ]
    )
    plate = PlateInfo.create(reader)
    assert plate and plate.barcode == "Plate 1"


def test_create_result() -> None:
    reader = get_reader_from_lines(
        [
            "Results",
            "1,2,,",
            "3,4,,",
            ",,,",
            "",
        ]
    )

    expected = [
        Result(col="A", row="01", value=1),
        Result(col="A", row="02", value=2),
        Result(col="B", row="01", value=3),
        Result(col="B", row="02", value=4),
    ]
    assert Result.create(reader) == expected

    # After no more results, return empty list.
    assert Result.create(reader) == []


def test_create_result_early_return() -> None:
    reader = get_reader_from_lines(
        [
            "Plate information",
            "1,2,,",
            "3,4,,",
        ]
    )
    assert Result.create(reader) == []


def test_create_plates() -> None:
    reader = get_reader_from_lines(
        [
            "Plate information",
            "Plate,Repeat,Barcode,Measured height,Chamber temperature at start,Measinfo,Kinetics,Measurement date,",
            "2,1,,1.1,14.5,De=2nd Ex=Top Em=Top Wdw=1 (14),0,10/13/2022 3:08:06 PM,",
            "",
            "Background information",
            "junk",
            "",
            "Calculated results",
            "junk",
            "",
            "Results",
            "6,,7,",
            ",,,",
            ",8,,",
        ]
    )

    expected = [
        Plate(
            plate_info=PlateInfo(
                number="2",
                barcode="Plate 2",
                emission_filter_id="2nd",
                measurement_time="10/13/2022 3:08:06 PM",
                measured_height=1.1,
                chamber_temperature_at_start=14.5,
            ),
            results=[
                Result(col="A", row="01", value=6),
                Result(col="A", row="03", value=7),
                Result(col="C", row="02", value=8),
            ],
        )
    ]

    assert Plate.create(reader) == expected


def test_create_basic_assay_info() -> None:
    reader = get_reader_from_lines(
        [
            "Basic assay information",
            "Assay ID: ,,,,3134",
            "Assay Started: ,,,,10/13/2022 3:06:23 PM",
            "Assay Finished: ,,,,10/13/2022 3:08:15 PM",
            "Assay Exported: ,,,,10/13/2022 3:08:15 PM",
            "Protocol ID: ,,,,100302",
            "Protocol Name: ,,,,HTRF LASER Eu 665/620",
            "Serial#: ,,,,='1050209'",
        ]
    )

    expected = BasicAssayInfo(
        protocol_id="100302",
        assay_id="3134",
    )

    assert BasicAssayInfo.create(reader) == expected


def test_create_basic_assay_info_fails() -> None:
    reader = get_reader_from_lines([])
    # TODO(brian): Improve src to throw AllotropeConversionError (or AllotropyError?)
    with pytest.raises(Exception, match="Did not find \\^Basic assay information"):
        BasicAssayInfo.create(reader)


def test_create_plate_type() -> None:
    reader = get_reader_from_lines(
        [
            "Protocol information",
            "Protocol:",
            "Protocol name,,,,protocol",
            "property,,,,1",
            "",
            "Plate type:",
            "Name of the plate type,,,,plate",
            "Number of the wells in the plate,,,,2",
        ]
    )

    expected = PlateType(number_of_wells=2)

    assert PlateType.create(reader) == expected


def test_create_plate_maps() -> None:
    reader = get_reader_from_lines(
        [
            "Platemap:",
            "Plate,,,,1",
            "Group,,,,1",
            "",
            ",01,02,03",
            "A,- ,'',",
            "B,nan,LH01,",
            "C,STD,,S01",
            "",
            "- - Undefined",
            "",
            "",
            "",
            "Plate,,,,2",
            "Group,,,,3",
            "",
            ",01,02,03,04",
            "A,,'',,BL...",
            "B,nan,nan,,",
            "C,,,,",
            "",
            "- - Undefined",
            "",
        ]
    )

    expected = {
        "1": PlateMap(
            plate_n="1",
            group_n="1",
            sample_role_type_mapping={
                "01": {
                    "A": SampleRoleType.undefined_sample_role,
                    "C": SampleRoleType.standard_sample_role,
                },
                "02": {"B": SampleRoleType.control_sample_role},
                "03": {"C": SampleRoleType.sample_role},
            },
        ),
        "2": PlateMap(
            plate_n="2",
            group_n="3",
            sample_role_type_mapping={
                "04": {"A": SampleRoleType.blank_role},
            },
        ),
    }

    assert create_plate_maps(reader) == expected


def test_create_labels() -> None:
    reader = get_reader_from_lines(
        [
            "Labels",
            "FAKE_INSTRUMENT,,,,4000013",
            "Exc. filter,,,,Filter_1",
            "Using of excitation filter,,,,Top",
            "2nd exc. filter,,,,N/A",
            "Ems. filter,,,,Filter_2",
            "2nd ems. filter,,,,N/A",
            "Measurement height,,,,5 mm",
            "Using of emission filter,,,,Top",
            "Number of flashes,,,,10",
            "Number of flashes integrated,,,,1",
            "Reference AD gain,,,,2",
            "Last edited,,,,4/6/2022 5:55:00 PM",
            "Factory preset,,,,No",
            "",
            "",
            "Filters:",
            "Filter_1,,,,102",
            "Description,,,,X485 CWL=485nm BW=14nm Tmin=60%",
            "Factory preset,,,,Yes",
            "",
            "",
            "Filter_2,,,,102",
            "Description,,,,X485 CWL=520nm BW=20nm Tmin=60%",
            "Factory preset,,,,Yes",
        ]
    )

    expected = Labels(
        label="FAKE_INSTRUMENT",
        excitation_filter=Filter(
            name="Filter_1",
            wavelength=485,
            bandwidth=14,
        ),
        emission_filters={
            "1st": Filter(
                name="Filter_2",
                wavelength=520,
                bandwidth=20,
            ),
            "2nd": None,
        },
        scan_position_setting=ScanPositionSettingPlateReader.top_scan_position__plate_reader_,
        number_of_flashes=10.0,
        detector_gain_setting="2",
    )

    assert Labels.create(reader) == expected


def test_create_instrument() -> None:
    reader = get_reader_from_lines(
        [
            "Instrument:",
            "Serial number,,,,1",
            "Nickname,,,,bobby",
        ]
    )

    expected = Instrument(
        serial_number="1",
        nickname="bobby",
    )

    assert Instrument.create(reader) == expected



from io import IOBase
from typing import Optional

from allotropy.allotrope.models.pcr_benchling_2023_09_qpcr import (
    BaselineCorrectedReporterDataCube,
    CalculatedDataDocumentItem,
    ContainerType,
    DataProcessingDocument,
    DataProcessingDocument1,
    DataSourceAggregateDocument,
    DataSourceDocumentItem,
    DataSystemDocument,
    DeviceControlAggregateDocument,
    DeviceControlDocumentItem,
    DeviceSystemDocument,
    ExperimentType,
    MeasurementAggregateDocument,
    MeasurementDocumentItem,
    MeltingCurveDataCube,
    Model,
    NormalizedReporterDataCube,
    PassiveReferenceDyeDataCube,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    QPCRAggregateDocument,
    QPCRDocumentItem,
    ReporterDyeDataCube,
    SampleDocument,
    TCalculatedDataAggregateDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TNullableQuantityValueUnitless,
    TQuantityValueNumber,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
    TDatacubeComponent,
    TDatacubeData,
    TDatacubeStructure,
)
from allotropy.constants import ASM_CONVERTER_NAME, ASM_CONVERTER_VERSION
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_builders import (
    DataBuilder,
)
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_structure import (
    Data,
    Well,
    WellItem,
)
from allotropy.parsers.lines_reader import LinesReader
from allotropy.parsers.vendor_parser import VendorParser


class AppBioQuantStudioParser(VendorParser):
    def _parse(self, raw_contents: IOBase, file_name: str) -> Model:
        reader = LinesReader(raw_contents)
        data = DataBuilder.build(reader)
        return self._get_model(data, file_name)

    def _get_model(self, data: Data, file_name: str) -> Model:
        return Model(
            qPCR_aggregate_document=QPCRAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    device_identifier=data.header.device_identifier,
                    model_number=data.header.model_number,
                    device_serial_number=data.header.device_serial_number,
                ),
                data_system_document=DataSystemDocument(
                    data_system_instance_identifier="localhost",
                    file_name=file_name,
                    UNC_path="",  # unknown
                    software_name="Thermo QuantStudio",
                    software_version="1.0",
                    ASM_converter_name=ASM_CONVERTER_NAME,
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                qPCR_document=[
                    QPCRDocumentItem(
                        analyst=data.header.analyst,
                        measurement_aggregate_document=MeasurementAggregateDocument(
                            experimental_data_identifier=data.header.experimental_data_identifier,
                            experiment_type=data.header.experiment_type,
                            container_type=ContainerType.qPCR_reaction_block,
                            plate_well_count=TQuantityValueNumber(
                                value=data.header.plate_well_count
                            ),
                            measurement_document=[
                                self.get_measurement_document_item(
                                    data, well, well_item
                                )
                                for well_item in well.items.values()
                            ],
                        ),
                        calculated_data_aggregate_document=self.get_inner_calculated_data_aggregate_document(
                            well
                        ),
                    )
                    for well in data.wells
                ],
                calculated_data_aggregate_document=self.get_outter_calculated_data_aggregate_document(
                    data
                ),
            )
        )

    def get_inner_calculated_data_aggregate_document(
        self, well: Well
    ) -> Optional[TCalculatedDataAggregateDocument]:
        if not well.calculated_document:
            return None

        return TCalculatedDataAggregateDocument(
            calculated_data_document=[
                CalculatedDataDocumentItem(
                    calculated_data_identifier=well.calculated_document.uuid,
                    data_source_aggregate_document=DataSourceAggregateDocument(
                        data_source_document=[
                            DataSourceDocumentItem(
                                data_source_identifier=data_source.reference.uuid,
                                data_source_feature=data_source.feature,
                            )
                            for data_source in well.calculated_document.data_sources
                        ],
                    ),
                    calculated_data_name=well.calculated_document.name,
                    calculated_data_description=None,
                    calculated_datum=TQuantityValueUnitless(
                        value=well.calculated_document.value
                    ),
                )
            ],
        )

    def get_outter_calculated_data_aggregate_document(
        self, data: Data
    ) -> TCalculatedDataAggregateDocument:
        if data.header.experiment_type in [
            ExperimentType.comparative_CT_qPCR_experiment,
            ExperimentType.relative_standard_curve_qPCR_experiment,
        ]:
            data_processing_document = DataProcessingDocument1(
                reference_DNA_description=data.endogenous_control,
                reference_sample_description=data.reference_sample,
            )
        else:
            data_processing_document = None

        return TCalculatedDataAggregateDocument(
            calculated_data_document=[
                CalculatedDataDocumentItem(
                    calculated_data_identifier=calc_doc.uuid,
                    data_source_aggregate_document=DataSourceAggregateDocument(
                        data_source_document=[
                            DataSourceDocumentItem(
                                data_source_identifier=(
                                    data_source.reference.uuid
                                    if data_source.reference
                                    else None
                                ),
                                data_source_feature=data_source.feature,
                            )
                            for data_source in calc_doc.data_sources
                        ],
                    ),
                    data_processing_document=data_processing_document,
                    calculated_data_name=calc_doc.name,
                    calculated_data_description=None,
                    calculated_datum=TQuantityValueUnitless(value=calc_doc.value),
                )
                for calc_doc in data.calculated_documents
            ],
        )

    def get_measurement_document_item(
        self, data: Data, well: Well, well_item: WellItem
    ) -> MeasurementDocumentItem:
        return MeasurementDocumentItem(
            measurement_identifier=well_item.uuid,
            measurement_time=self.get_date_time(data.header.measurement_time),
            target_DNA_description=well_item.target_dna_description,
            sample_document=SampleDocument(
                sample_identifier=well_item.sample_identifier,
                sample_role_type=well_item.sample_role_type,
                well_location_identifier=well_item.well_location_identifier,
                well_plate_identifier=data.header.barcode,
            ),
            device_control_aggregate_document=DeviceControlAggregateDocument(
                device_control_document=[
                    self.get_device_control_document_item(data, well_item),
                ],
            ),
            processed_data_aggregate_document=ProcessedDataAggregateDocument(
                processed_data_document=[
                    self.get_processed_data_document(well_item),
                ]
            ),
            reporter_dye_data_cube=self.get_reporter_dye_data_cube(well, well_item),
            passive_reference_dye_data_cube=self.get_passive_reference_dye_data_cube(
                data, well
            ),
            melting_curve_data_cube=self.get_melting_curve_data_cube(well),
        )

    def get_device_control_document_item(
        self, data: Data, well_item: WellItem
    ) -> DeviceControlDocumentItem:
        return DeviceControlDocumentItem(
            device_type="qPCR",
            measurement_method_identifier=data.header.measurement_method_identifier,
            total_cycle_number_setting=TQuantityValueNumber(
                value=well_item.amplification_data.total_cycle_number_setting,
            ),
            PCR_detection_chemistry=data.header.pcr_detection_chemistry,
            reporter_dye_setting=well_item.reporter_dye_setting,
            quencher_dye_setting=well_item.quencher_dye_setting,
            passive_reference_dye_setting=data.header.passive_reference_dye_setting,
        )

    def get_processed_data_document(
        self, well_item: WellItem
    ) -> ProcessedDataDocumentItem:
        normalized_reporter_result = (
            None
            if well_item.result.normalized_reporter_result is None
            else TQuantityValueUnitless(
                value=well_item.result.normalized_reporter_result
            )
        )

        baseline_corrected_reporter_result = (
            None
            if well_item.result.baseline_corrected_reporter_result is None
            else TQuantityValueUnitless(
                value=well_item.result.baseline_corrected_reporter_result
            )
        )

        genotyping_determination_method_setting = (
            None
            if well_item.result.genotyping_determination_method_setting is None
            else TQuantityValueUnitless(
                value=well_item.result.genotyping_determination_method_setting
            )
        )

        return ProcessedDataDocumentItem(
            data_processing_document=DataProcessingDocument(
                automatic_cycle_threshold_enabled_setting=well_item.result.automatic_baseline_determination_enabled_setting,
                cycle_threshold_value_setting=TQuantityValueUnitless(
                    value=well_item.result.cycle_threshold_value_setting,
                ),
                automatic_baseline_determination_enabled_setting=well_item.result.automatic_baseline_determination_enabled_setting,
                genotyping_determination_method_setting=genotyping_determination_method_setting,
            ),
            cycle_threshold_result=TNullableQuantityValueUnitless(
                value=well_item.result.cycle_threshold_result,
            ),
            normalized_reporter_result=normalized_reporter_result,
            baseline_corrected_reporter_result=baseline_corrected_reporter_result,
            genotyping_determination_result=well_item.result.genotyping_determination_result,
            normalized_reporter_data_cube=NormalizedReporterDataCube(
                label="normalized reporter",
                cube_structure=TDatacubeStructure(
                    dimensions=[
                        TDatacubeComponent(
                            field_componentDatatype=FieldComponentDatatype.integer,
                            concept="cycle count",
                            unit="#",
                        ),
                    ],
                    measures=[
                        TDatacubeComponent(
                            field_componentDatatype=FieldComponentDatatype.double,
                            concept="normalized report result",
                            unit="(unitless)",
                        ),
                    ],
                ),
                data=TDatacubeData(
                    dimensions=[well_item.amplification_data.cycle],
                    measures=[well_item.amplification_data.rn],
                ),
            ),
            baseline_corrected_reporter_data_cube=BaselineCorrectedReporterDataCube(
                label="baseline corrected reporter",
                cube_structure=TDatacubeStructure(
                    dimensions=[
                        TDatacubeComponent(
                            field_componentDatatype=FieldComponentDatatype.integer,
                            concept="cycle count",
                            unit="#",
                        ),
                    ],
                    measures=[
                        TDatacubeComponent(
                            field_componentDatatype=FieldComponentDatatype.double,
                            concept="baseline corrected reporter result",
                            unit="(unitless)",
                        ),
                    ],
                ),
                data=TDatacubeData(
                    dimensions=[well_item.amplification_data.cycle],
                    measures=[well_item.amplification_data.delta_rn],
                ),
            ),
        )

    def get_reporter_dye_data_cube(
        self, well: Well, well_item: WellItem
    ) -> Optional[ReporterDyeDataCube]:
        if well.multicomponent_data is None or well_item.reporter_dye_setting is None:
            return None

        return ReporterDyeDataCube(
            label="reporter dye",
            cube_structure=TDatacubeStructure(
                dimensions=[
                    TDatacubeComponent(
                        field_componentDatatype=FieldComponentDatatype.integer,
                        concept="cycle count",
                        unit="#",
                    ),
                ],
                measures=[
                    TDatacubeComponent(
                        field_componentDatatype=FieldComponentDatatype.double,
                        concept="reporter dye fluorescence",
                        unit="RFU",
                    ),
                ],
            ),
            data=TDatacubeData(
                dimensions=[well.multicomponent_data.cycle],
                measures=[
                    well.multicomponent_data.get_column(well_item.reporter_dye_setting)
                ],
            ),
        )

    def get_passive_reference_dye_data_cube(
        self, data: Data, well: Well
    ) -> Optional[PassiveReferenceDyeDataCube]:
        if (
            well.multicomponent_data is None
            or data.header.passive_reference_dye_setting is None
        ):
            return None

        return PassiveReferenceDyeDataCube(
            label="passive reference dye",
            cube_structure=TDatacubeStructure(
                dimensions=[
                    TDatacubeComponent(
                        field_componentDatatype=FieldComponentDatatype.integer,
                        concept="cycle count",
                        unit="#",
                    ),
                ],
                measures=[
                    TDatacubeComponent(
                        field_componentDatatype=FieldComponentDatatype.double,
                        concept="passive reference dye fluorescence",
                        unit="RFU",
                    ),
                ],
            ),
            data=TDatacubeData(
                dimensions=[well.multicomponent_data.cycle],
                measures=[
                    well.multicomponent_data.get_column(
                        data.header.passive_reference_dye_setting
                    )
                ],
            ),
        )

    def get_melting_curve_data_cube(self, well: Well) -> Optional[MeltingCurveDataCube]:
        if well.melt_curve_raw_data is None:
            return None

        return MeltingCurveDataCube(
            label="melting curve",
            cube_structure=TDatacubeStructure(
                dimensions=[
                    TDatacubeComponent(
                        field_componentDatatype=FieldComponentDatatype.double,
                        concept="temperature",
                        unit="degrees C",
                    ),
                ],
                measures=[
                    TDatacubeComponent(
                        field_componentDatatype=FieldComponentDatatype.double,
                        concept="reporter dye fluorescence",
                        unit="(unitless)",
                    ),
                    TDatacubeComponent(
                        field_componentDatatype=FieldComponentDatatype.double,
                        concept="slope",
                        unit="(unitless)",
                    ),
                ],
            ),
            data=TDatacubeData(
                dimensions=[well.melt_curve_raw_data.reading],
                measures=[
                    well.melt_curve_raw_data.fluorescence,
                    well.melt_curve_raw_data.derivative,
                ],
            ),
        )



from io import StringIO
from typing import Any, Optional
import uuid

import numpy as np
import pandas as pd

from allotropy.allotrope.allotrope import AllotropeConversionError
from allotropy.allotrope.models.pcr_benchling_2023_09_qpcr import (
    ExperimentType,
)
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_calculated_documents import (
    build_quantity,
    iter_calculated_data_documents,
)
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_structure import (
    AmplificationData,
    Data,
    Header,
    MeltCurveRawData,
    MulticomponentData,
    RawData,
    Result,
    Well,
    WellItem,
    WellList,
)
from allotropy.parsers.lines_reader import LinesReader


def df_to_series(df: pd.DataFrame) -> pd.Series:
    return pd.Series(df.iloc[0], index=df.columns)


def float_or_none(value: Any) -> Optional[float]:
    try:
        return float(value)
    except Exception:
        return None


def get_str(data: pd.Series, key: str, default: Optional[str] = None) -> Optional[str]:
    value = data.get(key, default)
    return None if value is None else str(value)


def get_int(data: pd.Series, key: str) -> Optional[int]:
    try:
        value = data.get(key)
        return None if value is None else int(value)  # type: ignore[arg-type]
    except Exception as e:
        msg = f"Unable to convert {key} to integer value"
        raise AllotropeConversionError(msg) from e


def get_float(data: pd.Series, key: str) -> Optional[float]:
    try:
        value = data.get(key)
        return float_or_none(value)
    except Exception as e:
        msg = f"Unable to convert {key} to float value"
        raise AllotropeConversionError(msg) from e


def get_bool(data: pd.Series, key: str) -> Optional[bool]:
    try:
        value = data.get(key)
        return None if value is None else bool(value)
    except Exception as e:
        msg = f"Unable to convert {key} to bool value"
        raise AllotropeConversionError(msg) from e


class HeaderBuilder:
    @staticmethod
    def build(reader: LinesReader) -> Header:
        data = HeaderBuilder.get_data(reader)

        model_number = get_str(data, "Instrument Type")
        if model_number is None:
            msg = "Unable to get model number"
            raise AllotropeConversionError(msg)

        measurement_method_identifier = get_str(data, "Quantification Cycle Method")
        if measurement_method_identifier is None:
            msg = "Unable to get measurement method identifier"
            raise AllotropeConversionError(msg)

        pcr_detection_chemistry = get_str(data, "Chemistry")
        if pcr_detection_chemistry is None:
            msg = "Unable to get PCR detection chemistry"
            raise AllotropeConversionError(msg)

        return Header(
            measurement_time=HeaderBuilder.get_measurement_time(data),
            plate_well_count=HeaderBuilder.get_plate_well_count(data),
            experiment_type=HeaderBuilder.get_experiment_type(data),
            device_identifier=get_str(data, "Instrument Name") or "NA",
            model_number=model_number,
            device_serial_number=get_str(data, "Instrument Serial Number") or "NA",
            measurement_method_identifier=measurement_method_identifier,
            pcr_detection_chemistry=pcr_detection_chemistry,
            passive_reference_dye_setting=get_str(data, "Passive Reference"),
            barcode=get_str(data, "Experiment Barcode"),
            analyst=get_str(data, "Experiment User Name"),
            experimental_data_identifier=get_str(data, "Experiment Name"),
        )

    @staticmethod
    def get_experiment_type(data: pd.Series) -> ExperimentType:
        experiments_type_options = {
            "Standard Curve": ExperimentType.standard_curve_qPCR_experiment,
            "Relative Standard Curve": ExperimentType.relative_standard_curve_qPCR_experiment,
            "Comparative Cт (ΔΔCт)": ExperimentType.comparative_CT_qPCR_experiment,
            "Melt Curve": ExperimentType.melt_curve_qPCR_experiment,
            "Genotyping": ExperimentType.genotyping_qPCR_experiment,
            "Presence/Absence": ExperimentType.presence_absence_qPCR_experiment,
        }

        experiments_type = experiments_type_options.get(data.get("Experiment Type"))  # type: ignore[arg-type]
        if experiments_type is None:
            msg = "Unable to get valid experiment type"
            raise AllotropeConversionError(msg)
        return experiments_type

    @staticmethod
    def get_data(reader: LinesReader) -> pd.Series:
        lines = [line.replace("*", "", 1) for line in reader.pop_until(r"^\[.+\]")]
        csv_stream = StringIO("\n".join(lines))
        raw_data = pd.read_csv(
            csv_stream, header=None, sep="=", names=["index", "values"]
        )
        data = pd.Series(raw_data["values"].values, index=raw_data["index"])
        data.index = data.index.str.strip()
        return data.str.strip().replace("NA", None)

    @staticmethod
    def get_measurement_time(data: pd.Series) -> str:
        return str(data.get("Experiment Run End Time"))

    @staticmethod
    def get_plate_well_count(data: pd.Series) -> int:
        block_type = get_str(data, "Block Type", default="")

        if block_type is None:
            msg = "Unable to get plate well count"
            raise AllotropeConversionError(msg)

        if "96" in block_type:
            return 96
        elif "384" in block_type:
            return 384

        msg = "Unable to interpret plate well count"
        raise AllotropeConversionError(msg)


class GenericWellBuilder:
    @staticmethod
    def build(data: pd.DataFrame) -> WellList:
        return WellList(
            [
                Well(
                    identifier=identifier,  # type: ignore[arg-type]
                    items={
                        item_data["Target Name"]: GenericWellBuilder.build_well_item(
                            item_data
                        )
                        for _, item_data in well_data.iterrows()
                    },
                )
                for identifier, well_data in data.groupby("Well")
            ]
        )

    @staticmethod
    def build_well_item(data: pd.Series) -> WellItem:
        identifier = get_int(data, "Well")
        if identifier is None:
            msg = "Unable to get well identifier"
            raise AllotropeConversionError(msg)

        target_dna_description = get_str(data, "Target Name")
        if target_dna_description is None:
            msg = f"Unable to get target dna description for well {identifier}"
            raise AllotropeConversionError(msg)

        sample_identifier = get_str(data, "Sample Name")
        if sample_identifier is None:
            msg = f"Unable to get sample identifier for well {identifier}"
            raise AllotropeConversionError(msg)

        return WellItem(
            uuid=str(uuid.uuid4()),
            identifier=identifier,
            target_dna_description=target_dna_description,
            sample_identifier=sample_identifier,
            reporter_dye_setting=get_str(data, "Reporter"),
            position=get_str(data, "Well Position", default="UNDEFINED"),
            well_location_identifier=get_str(data, "Well Position"),
            quencher_dye_setting=get_str(data, "Quencher"),
            sample_role_type=get_str(data, "Task"),
        )


class GenotypingWellBuilder:
    @staticmethod
    def build(data: pd.DataFrame) -> WellList:
        return WellList(
            [
                Well(
                    identifier=well_id,  # type: ignore[arg-type]
                    items={
                        well_item.target_dna_description: well_item
                        for well_item in GenotypingWellBuilder.build_well_items(
                            well_data
                        )
                    },
                )
                for well_id, well_data in data.iterrows()
            ]
        )

    @staticmethod
    def build_well_items(data: pd.Series) -> list[WellItem]:
        identifier = get_int(data, "Well")
        if identifier is None:
            msg = "Unable to get well identifier"
            raise AllotropeConversionError(msg)

        snp_name = get_str(data, "SNP Assay Name")
        if snp_name is None:
            msg = f"Unable to get snp name for well {identifier}"
            raise AllotropeConversionError(msg)

        allele1 = get_str(data, "Allele1 Name")
        if allele1 is None:
            msg = f"Unable to get allele 1 for well {identifier}"
            raise AllotropeConversionError(msg)

        allele2 = get_str(data, "Allele2 Name")
        if allele2 is None:
            msg = f"Unable to get allele 2 for well {identifier}"
            raise AllotropeConversionError(msg)

        sample_identifier = get_str(data, "Sample Name")
        if sample_identifier is None:
            msg = "Unable to get sample identifier"
            raise AllotropeConversionError(msg)

        return [
            WellItem(
                uuid=str(uuid.uuid4()),
                identifier=identifier,
                target_dna_description=f"{snp_name}-{allele}",
                sample_identifier=sample_identifier,
                reporter_dye_setting=get_str(data, "Reporter"),
                position=get_str(data, "Well Position", default="UNDEFINED"),
                well_location_identifier=get_str(data, "Well Position"),
                quencher_dye_setting=get_str(data, "Quencher"),
                sample_role_type=get_str(data, "Task"),
            )
            for allele in [allele1, allele2]
        ]


class WellBuilder:
    @staticmethod
    def build(reader: LinesReader, experiment_type: ExperimentType) -> WellList:
        raw_data = WellBuilder.get_data(reader)
        data = raw_data[raw_data["Sample Name"].notnull()]
        if experiment_type == ExperimentType.genotyping_qPCR_experiment:
            return GenotypingWellBuilder.build(data)
        return GenericWellBuilder.build(data)

    @staticmethod
    def get_data(reader: LinesReader) -> pd.DataFrame:
        if reader.drop_until(r"^\[Sample Setup\]") is None:
            msg = "Unable to find Sample Setup section in input file"
            raise AllotropeConversionError(msg)

        reader.pop()  # remove title
        lines = list(reader.pop_until(r"^\[.+\]"))
        csv_stream = StringIO("\n".join(lines))
        return pd.read_csv(csv_stream, sep="\t").replace(np.nan, None)


class RawDataBuilder:
    @staticmethod
    def build(reader: LinesReader) -> Optional[RawData]:
        if not reader.match(r"^\[Raw Data\]"):
            return None
        reader.pop()  # remove title
        lines = list(reader.pop_until(r"^\[.+\]"))
        return RawData(lines)


class AmplificationDataBuilder:
    @staticmethod
    def build(
        amplification_data: pd.DataFrame, well_item: WellItem
    ) -> AmplificationData:
        target_data = AmplificationDataBuilder.filter_target_data(
            amplification_data, well_item
        )
        return AmplificationData(
            total_cycle_number_setting=float(target_data["Cycle"].max()),
            cycle=target_data["Cycle"].tolist(),
            rn=target_data["Rn"].tolist(),
            delta_rn=target_data["Delta Rn"].tolist(),
        )

    @staticmethod
    def filter_target_data(
        amplification_data: pd.DataFrame, well_item: WellItem
    ) -> pd.DataFrame:
        well_data = amplification_data[
            amplification_data["Well"] == well_item.identifier
        ]
        if well_data.empty:
            msg = f"Unable to get amplification data for well {well_item.identifier}"
            raise AllotropeConversionError(msg)

        target_data = well_data[
            well_data["Target Name"] == well_item.target_dna_description
        ]
        if target_data.empty:
            msg = f"Unable to get amplification data for well {well_item.identifier}"
            raise AllotropeConversionError(msg)

        return target_data

    @staticmethod
    def get_data(reader: LinesReader) -> pd.DataFrame:
        if reader.drop_until(r"^\[Amplification Data\]") is None:
            msg = "Unable to find Amplification Data section in input file"
            raise AllotropeConversionError(msg)

        reader.pop()  # remove title
        lines = list(reader.pop_until(r"^\[.+\]"))
        csv_stream = StringIO("\n".join(lines))
        return pd.read_csv(csv_stream, sep="\t", thousands=r",")


class MulticomponentDataBuilder:
    @staticmethod
    def build(data: pd.DataFrame, well: Well) -> MulticomponentData:
        well_data = MulticomponentDataBuilder.filter_well_data(data, well)
        return MulticomponentData(
            cycle=well_data["Cycle"].tolist(),
            columns={
                name: well_data[name].tolist()  # type: ignore[misc]
                for name in well_data
                if name not in ["Well", "Cycle", "Well Position"]
            },
        )

    @staticmethod
    def filter_well_data(data: pd.DataFrame, well: Well) -> pd.DataFrame:
        well_data = data[data["Well"] == well.identifier]
        if well_data.empty:
            msg = f"Unable to find multi component data for well {well.identifier}"
            raise AllotropeConversionError(msg)

        return well_data

    @staticmethod
    def get_data(reader: LinesReader) -> Optional[pd.DataFrame]:
        if not reader.match(r"^\[Multicomponent Data\]"):
            return None
        reader.pop()  # remove title
        lines = list(reader.pop_until(r"^\[.+\]"))
        csv_stream = StringIO("\n".join(lines))
        return pd.read_csv(csv_stream, sep="\t", thousands=r",")


class GenericResultsBuilder:
    @staticmethod
    def build(data: pd.DataFrame, well_item: WellItem) -> Result:
        target_data = GenericResultsBuilder.filter_target_data(data, well_item)
        cycle_threshold_value_setting = get_float(target_data, "Ct Threshold")
        if cycle_threshold_value_setting is None:
            msg = "Unable to get cycle threshold value setting"
            raise AllotropeConversionError(msg)

        cycle_threshold_result = target_data.get("CT")
        if cycle_threshold_result is None:
            msg = "Unable to get cycle threshold result"
            raise AllotropeConversionError(msg)

        return Result(
            cycle_threshold_value_setting=cycle_threshold_value_setting,
            cycle_threshold_result=float_or_none(cycle_threshold_result),
            automatic_cycle_threshold_enabled_setting=get_bool(
                target_data, "Automatic Ct Threshold"
            ),
            automatic_baseline_determination_enabled_setting=get_bool(
                target_data, "Automatic Baseline"
            ),
            normalized_reporter_result=get_float(target_data, "Rn"),
            baseline_corrected_reporter_result=get_float(target_data, "Delta Rn"),
            genotyping_determination_result=get_str(target_data, "Call"),
            genotyping_determination_method_setting=get_float(
                target_data, "Threshold Value"
            ),
            quantity=get_float(target_data, "Quantity"),
            quantity_mean=get_float(target_data, "Quantity Mean"),
            quantity_sd=get_float(target_data, "Quantity SD"),
            ct_mean=get_float(target_data, "Ct Mean"),
            ct_sd=get_float(target_data, "Ct SD"),
            delta_ct_mean=get_float(target_data, "Delta Ct Mean"),
            delta_ct_se=get_float(target_data, "Delta Ct SE"),
            delta_delta_ct=get_float(target_data, "Delta Delta Ct"),
            rq=get_float(target_data, "RQ"),
            rq_min=get_float(target_data, "RQ Min"),
            rq_max=get_float(target_data, "RQ Max"),
            rn_mean=get_float(target_data, "Rn Mean"),
            rn_sd=get_float(target_data, "Rn SD"),
            y_intercept=get_float(target_data, "Y-Intercept"),
            r_squared=get_float(target_data, "R(superscript 2)"),
            slope=get_float(target_data, "Slope"),
            efficiency=get_float(target_data, "Efficiency"),
        )

    @staticmethod
    def filter_target_data(data: pd.DataFrame, well_item: WellItem) -> pd.Series:
        well_data = data[data["Well"] == well_item.identifier]
        if well_data.empty:
            msg = f"Unable to get result data for well {well_item.identifier}"
            raise AllotropeConversionError(msg)

        target_data = well_data[
            well_data["Target Name"] == well_item.target_dna_description
        ]
        if target_data.empty:
            msg = f"Unable to get result data for well {well_item.identifier}"
            raise AllotropeConversionError(msg)

        n_rows, _ = target_data.shape
        if n_rows != 1:
            msg = f"Unexpected number of results associated to well {well_item.identifier}"
            raise AllotropeConversionError(msg)

        return df_to_series(target_data)


class GenotypingResultsBuilder:
    @staticmethod
    def build(data: pd.DataFrame, well_item: WellItem) -> Result:
        target_data = GenotypingResultsBuilder.filter_target_data(data, well_item)
        _, raw_allele = well_item.target_dna_description.split("-")
        allele = raw_allele.replace(" ", "")
        cycle_threshold_value_setting = get_float(target_data, f"{allele} Ct Threshold")
        if cycle_threshold_value_setting is None:
            msg = f"Unable to get cycle threshold value setting for well {well_item.identifier}"
            raise AllotropeConversionError(msg)

        cycle_threshold_result = target_data.get(f"{allele} Ct")
        if cycle_threshold_result is None:
            msg = "Unable to get cycle threshold result"
            raise AllotropeConversionError(msg)

        return Result(
            cycle_threshold_value_setting=cycle_threshold_value_setting,
            cycle_threshold_result=float_or_none(cycle_threshold_result),
            automatic_cycle_threshold_enabled_setting=get_bool(
                target_data, f"{allele} Automatic Ct Threshold"
            ),
            automatic_baseline_determination_enabled_setting=get_bool(
                target_data, f"{allele} Automatic Baseline"
            ),
            normalized_reporter_result=get_float(target_data, "Rn"),
            baseline_corrected_reporter_result=get_float(
                target_data, f"{allele} Delta Rn"
            ),
            genotyping_determination_result=get_str(target_data, "Call"),
            genotyping_determination_method_setting=get_float(
                target_data, "Threshold Value"
            ),
            quantity=get_float(target_data, "Quantity"),
            quantity_mean=get_float(target_data, "Quantity Mean"),
            quantity_sd=get_float(target_data, "Quantity SD"),
            ct_mean=get_float(target_data, "Ct Mean"),
            ct_sd=get_float(target_data, "Ct SD"),
            delta_ct_mean=get_float(target_data, "Delta Ct Mean"),
            delta_ct_se=get_float(target_data, "Delta Ct SE"),
            delta_delta_ct=get_float(target_data, "Delta Delta Ct"),
            rq=get_float(target_data, "RQ"),
            rq_min=get_float(target_data, "RQ Min"),
            rq_max=get_float(target_data, "RQ Max"),
            rn_mean=get_float(target_data, "Rn Mean"),
            rn_sd=get_float(target_data, "Rn SD"),
            y_intercept=get_float(target_data, "Y-Intercept"),
            r_squared=get_float(target_data, "R(superscript 2)"),
            slope=get_float(target_data, "Slope"),
            efficiency=get_float(target_data, "Efficiency"),
        )

    @staticmethod
    def filter_target_data(data: pd.DataFrame, well_item: WellItem) -> pd.Series:
        well_data = data[data["Well"] == well_item.identifier]
        if well_data.empty:
            msg = f"Unable to get result data for well {well_item.identifier}"
            raise AllotropeConversionError(msg)

        snp_assay_name, _ = well_item.target_dna_description.split("-")
        target_data = well_data[well_data["SNP Assay Name"] == snp_assay_name]
        if target_data.empty:
            msg = f"Unable to get result data for well {well_item.identifier}"
            raise AllotropeConversionError(msg)

        n_rows, _ = target_data.shape
        if n_rows != 1:
            msg = f"Unexpected number of results associated to well {well_item.identifier}"
            raise AllotropeConversionError(msg)

        return df_to_series(target_data)


class ResultsBuilder:
    @staticmethod
    def build(
        data: pd.DataFrame, well_item: WellItem, experiment_type: ExperimentType
    ) -> Result:
        if experiment_type == ExperimentType.genotyping_qPCR_experiment:
            return GenotypingResultsBuilder.build(data, well_item)
        return GenericResultsBuilder.build(data, well_item)

    @staticmethod
    def get_data(reader: LinesReader) -> tuple[pd.DataFrame, pd.Series]:
        if reader.drop_until(r"^\[Results\]") is None:
            msg = "Unable to find Results section in input file"
            raise AllotropeConversionError(msg)

        reader.pop()  # remove title
        data_lines = list(reader.pop_until_empty())
        csv_stream = StringIO("\n".join(data_lines))
        data = pd.read_csv(csv_stream, sep="\t", thousands=r",").replace(np.nan, None)

        reader.drop_empty()

        if reader.match(r"\[.+\]"):
            return data, pd.Series()

        metadata_lines = list(reader.pop_until_empty())
        csv_stream = StringIO("\n".join(metadata_lines))
        raw_data = pd.read_csv(
            csv_stream, header=None, sep="=", names=["index", "values"]
        )
        metadata = pd.Series(raw_data["values"].values, index=raw_data["index"])
        metadata.index = metadata.index.str.strip()

        reader.drop_empty()

        return data, metadata.str.strip()


class MeltCurveRawDataBuilder:
    @staticmethod
    def build(data: pd.DataFrame, well: Well) -> MeltCurveRawData:
        well_data = MeltCurveRawDataBuilder.filter_well_data(data, well)
        return MeltCurveRawData(
            reading=well_data["Reading"].tolist(),
            fluorescence=well_data["Fluorescence"].tolist(),
            derivative=well_data["Derivative"].tolist(),
        )

    @staticmethod
    def filter_well_data(data: pd.DataFrame, well: Well) -> pd.DataFrame:
        well_data = data[data["Well"] == well.identifier]
        if well_data.empty:
            msg = f"Unable to get melt curve raw data for well {well.identifier}"
            raise AllotropeConversionError(msg)

        return well_data

    @staticmethod
    def get_data(reader: LinesReader) -> Optional[pd.DataFrame]:
        if not reader.match(r"^\[Melt Curve Raw Data\]"):
            return None
        reader.pop()  # remove title
        lines = list(reader.pop_until_empty())
        csv_stream = StringIO("\n".join(lines))
        return pd.read_csv(csv_stream, sep="\t", thousands=r",")


class DataBuilder:
    @staticmethod
    def build(reader: LinesReader) -> Data:
        header = HeaderBuilder.build(reader)
        wells = WellBuilder.build(reader, header.experiment_type)
        raw_data = RawDataBuilder.build(reader)

        amp_data = AmplificationDataBuilder.get_data(reader)
        multi_data = MulticomponentDataBuilder.get_data(reader)
        results_data, results_metadata = ResultsBuilder.get_data(reader)
        melt_data = MeltCurveRawDataBuilder.get_data(reader)
        for well in wells:
            if multi_data is not None:
                well.add_multicomponent_data(
                    MulticomponentDataBuilder.build(multi_data, well)
                )
            if melt_data is not None:
                well.add_melt_curve_raw_data(
                    MeltCurveRawDataBuilder.build(melt_data, well)
                )
            for well_item in well.items.values():
                well_item.add_amplification_data(
                    AmplificationDataBuilder.build(amp_data, well_item)
                )
                well_item.add_result(
                    ResultsBuilder.build(
                        results_data, well_item, header.experiment_type
                    )
                )

            if an_well_item := well.get_an_well_item():
                well.calculated_document = build_quantity(an_well_item)

        endogenous_control = get_str(results_metadata, "Endogenous Control") or ""
        reference_sample = get_str(results_metadata, "Reference Sample") or ""

        return Data(
            header,
            wells,
            raw_data,
            endogenous_control,
            reference_sample,
            list(
                iter_calculated_data_documents(
                    wells,
                    header.experiment_type,
                    reference_sample,
                    endogenous_control,
                )
            ),
        )



import re
from typing import Optional, TypeVar, Union

from allotropy.allotrope.allotrope import AllotropeConversionError

PrimitiveValue = Union[str, int, float]


def try_int(value: Optional[str]) -> Optional[int]:
    try:
        return int(value or "")
    except ValueError:
        return None


def try_float(value: Optional[str]) -> Optional[float]:
    try:
        return float(value or "")
    except ValueError:
        return None


def natural_sort_key(key: str) -> list[str]:
    """Returns a sort key that treats numeric substrings as parsed integers for comparison."""
    tokens = [token for token in re.split(r"(\d+)", key) if token]
    return [
        f"{int(token):>10}" if token.isdecimal() else token.lower() for token in tokens
    ]


T = TypeVar("T")


def assert_not_none(
    value: Optional[T], name: Optional[str] = None, msg: Optional[str] = None
) -> T:
    if value is None:
        error = msg or f"Expected non-null value{f' for {name}' if name else ''}"
        raise AllotropeConversionError(error)
    return value


def value_or_none(value: str) -> Optional[str]:
    return value.strip() or None
